from meteor_reasoner.demo.dataset import CASES
from meteor_reasoner.utils.loader import load_program
from meteor_reasoner.canonical.calculate_w import get_w


def test_w():
    case = CASES["case_1"]
    program = case[1]
    truth_w = case[2][0]

    Program = load_program(program)
    w = get_w(rules=Program)
    print("w:",w)
    print("truth_w:",truth_w)
    assert 2 * w == truth_w

    case = CASES["case_2"]
    program = case[1]
    truth_w = case[2][0]

    Program = load_program(program)
    w = get_w(rules=Program)
    print("w:", w)
    print("truth_w:", truth_w)
    assert 2 * w == truth_w

    case = CASES["case_3"]
    program = case[1]
    truth_w = case[2][0]

    Program = load_program(program)
    w = get_w(rules=Program)
    print("w:", w)
    print("truth_w:", truth_w)
    assert 2 * w == truth_w

    case = CASES["case_4"]
    program = case[1]
    truth_w = case[2][0]

    Program = load_program(program)
    w = get_w(rules=Program)
    print("w:", w)
    print("truth_w:", truth_w)
    assert 2 * w == truth_w

    case = CASES["case_5"]
    program = case[1]
    truth_w = case[2][0]

    Program = load_program(program)
    w = get_w(rules=Program)
    print("w:", w)
    print("truth_w:", truth_w)
    assert 2 * w == truth_w

    case = CASES["case_6"]
    program = case[1]
    truth_w = case[2][0]

    Program = load_program(program)
    w = get_w(rules=Program)
    print("w:", w)
    print("truth_w:", truth_w)
    assert 2 * w == truth_w

    case = CASES["case_7"]
    program = case[1]
    truth_w = case[2][0]

    Program = load_program(program)
    w = get_w(rules=Program)
    print("w:", w)
    print("truth_w:", truth_w)
    assert 2 * w == truth_w

    case = CASES["case_8"]
    program = case[1]
    truth_w = case[2][0]

    Program = load_program(program)
    w = get_w(rules=Program)
    print("w:", w)
    print("truth_w:", truth_w)
    assert 2 * w == truth_w

test_w()






class Rule:
    """
    A rule instance consists of two part: a head (an Atom instance) and
    a body (a list of  Literal or BinarayLiteral Instances)
    """
    def __init__(self, head, body, negative_body=[]):
        """

        Args:
            head (an Atom instance):
            body (a list of Literal or BinaryLiteral instances):
        """
        self.head = head
        self.body = body  # a list of literal
        self.negative_body = negative_body # a list of literals

    def __str__(self):
        if len(self.negative_body) == 0:
            return str(self.head) + ":-" + ",".join([str(literal) for literal in self.body])
        else:
            return str(self.head) + ":-" + ",".join([str(literal) for literal in self.body])\
                   + "||" + ",".join([str(literal) for literal in self.negative_body_atoms])


from meteor_reasoner.materialization.coalesce import *
from meteor_reasoner.materialization.index_build import *
from meteor_reasoner.utils.loader import *
from meteor_reasoner.utils.operate_dataset import print_dataset, print_predicate
from meteor_reasoner.materialization.materialize import materialize
import time
import logging
import sys
import argparse
parser = argparse.ArgumentParser()
parser.add_argument("--datapath", default="../datasets/ISWC/lubm_sample.txt",  type=str, help="Input the dataset path (json format)")
parser.add_argument("--rulepath", default="./programs/p1.txt", type=str, help="Input the program path")
parser.add_argument("--predicate", default="a1:GoodUniversity", type=str, help="Input the predicate you want to print")
parser.add_argument("--K", default=1, type=int, help="Default number of iterations")
parser.add_argument("--logpath", required=True, type=str, help="The name of logpath of the form: ./log/dataset_approach_program.txt")
args = parser.parse_args()

logger = logging.getLogger()
logger.setLevel(logging.INFO)

output_file_handler = logging.FileHandler(args.logpath)
stdout_handler = logging.StreamHandler(sys.stdout)
logger.addHandler(output_file_handler)
logger.addHandler(stdout_handler)


start_time = time.time()
with open(args.rulepath) as file:
    raw_program = file.readlines()
    program = load_program(raw_program)
    predicates = set()
    for rule in program:
        predicates.add(rule.head.get_predicate())
        for literal in rule.body:
            if isinstance(literal, BinaryLiteral):
                predicates.add(literal.left_literal.get_predicate())
                predicates.add(literal.right_literal.get_predicate())
            else:
                predicates.add(literal.get_predicate())

with open(args.datapath) as file:
    raw_data = []
    for line in file:
        raw_data.append(line)
        # for predicate in predicates:
        #     if line.startswith(predicate):
        #          raw_data.append(line)
        #          break
    D = load_dataset(raw_data)
    coalescing_d(D)
    D_index = build_index(D)

logger.info("Size of Dataset({}): {}".format(args.datapath, len(raw_data)))
start_time = time.time()
materialize(D, rules=program, D_index=D_index, seminaive=True, delta_old=D, K=args.K, logger=logger)
logger.info("======END=========\n")



from meteor_reasoner.graphutil.graph import *
from meteor_reasoner.materialization.ground import ground_generator
from meteor_reasoner.materialization.apply import apply


def no_new_facts(delta_new, D, limit):
    for predicate in delta_new:
        for entity in delta_new[predicate]:
            if predicate in D and entity in D[predicate]:
                diff_interval_list = Interval.diff_list(delta_new[predicate][entity], D[predicate][entity])
                if len(diff_interval_list) > 0:
                    for interval in diff_interval_list:
                        new_interval = Interval.intersection(limit, interval)
                        if new_interval is not None:
                            #  print("diff:", [str(item) for item  in diff_interval_list])
                            #  print("D:", [str(item) for item  in D[predicate][entity]])
                            #  print("new:", [str(item) for item  in delta_new[predicate][entity]])
                            #  print(predicate, [item.name for item in entity], str(interval))
                             return False
            else:
                for interval in delta_new[predicate][entity]:
                    new_interval = Interval.intersection(limit, interval)
                    if new_interval is not None:
                        #  print("D:", [str(item) for item  in D[predicate][entity]])
                        #  print("new:", [str(item) for item  in delta_new[predicate][entity]])
                        #  print(predicate, [item.name for item in entity], str(interval))
                         return False
    return True


def literal_contain_no_variable(literal):
    if isinstance(literal, Atom):
        if literal.get_predicate() not in ["Bottom", "Top"] and len(literal.entity) == 1 and literal.entity[0].name == "nan":
            return True
    elif isinstance(literal, Literal):
        if isinstance(literal, Literal) and len(literal.atom.entity) == 1 and literal.atom.entity[0].name == "nan":
            return True
    elif isinstance(literal, BinaryLiteral):
        if literal.left_literal.get_predicate() in ["Bottom", "Top"] or (
                len(literal.left_literal.entity) == 1 and literal.left_literal.entity[0].name == "nan"):
            if literal.right_literal.get_predicate() in ["Bottom", "Top"] or (
                    len(literal.right_literal.entity) == 1 and literal.right_literal.entity[0].name == "nan"):
                return True


def split_rules_predicates(program):
    """
    This function split all rules in the given program into non-recursive rules and recursive rules.
    Besides, this function also split all predicates in the given program into involved predicates and automata predicated.

    None-recursive rules: rules needed to be completely materialized before running the automata
    Recursive rules: rules that can not be completely materialized
    Involved predicates refer to invariable predicates but are needed to be considered in the Automata construction
    Automata predicates refer to variable predicates in the Automata construction whose states are uncertain
    Args:
        program (list of Rule instances):

    Returns:
        set, set, set, set

    """
    graph = Graph(program)

    non_recursive_rules, automata_rules = graph.split_rules("a1:FullProfessor")
    involved_predicates = set()
    automata_predicates = set()

    for rule in non_recursive_rules:
        involved_predicates.add(rule.head.get_predicate())
        for literal in rule.body:
            if isinstance(literal, BinaryLiteral):
                if literal.left_literal.get_predicate() not in ["Bottom", "Top"]:
                    involved_predicates.add(literal.left_literal.get_predicate())
                if literal.right_literal.get_predicate() not in ["Bottom", "Top"]:
                    involved_predicates.add(literal.right_literal.get_predicate())
            else:
                if literal.get_predicate() not in ["Bottom", "Top"]:
                    involved_predicates.add(literal.get_predicate())

    for rule in automata_rules:
        automata_predicates.add(rule.head.get_predicate())
        involved_predicates.add(rule.head.get_predicate())
        for literal in rule.body:
            if isinstance(literal, BinaryLiteral):
                if literal.left_literal.get_predicate() not in ["Bottom", "Top"]:
                    involved_predicates.add(literal.left_literal.get_predicate())
                    automata_predicates.add(rule.head.get_predicate())
                if literal.right_literal.get_predicate() not in ["Bottom", "Top"]:
                    involved_predicates.add(literal.right_literal.get_predicate())
                    automata_predicates.add(rule.head.get_predicate())
            else:
                if literal.get_predicate() not in ["Bottom", "Top"]:
                    involved_predicates.add(literal.get_predicate())
                    automata_predicates.add(rule.head.get_predicate())

    return (non_recursive_rules, automata_rules, involved_predicates, automata_predicates)


def entail_same_nonrecursive_predicates(D, delta_new, non_predicates):
    for predicate in delta_new:
        if predicate in non_predicates:
            if predicate not in D:
                 return False
            for entity in delta_new[predicate]:
                   if entity in D[predicate]:
                       if not Interval.list_inclusion(delta_new[predicate][entity], D[predicate][entity]):
                            return False
                   else:
                       return False
    return True


def pre_calculate_threshold(rules, propagation, D, D_index, non_predicates):
    observed_rules = defaultdict(list)
    deleted_rules = set()
    if propagation == 1:
        for rule in rules:
            min_right = float("inf")
            contain_nrpredicate_flag = False
            for tmp_literal in rule.body:
                literal = copy.deepcopy(tmp_literal)
                intervals = []
                if isinstance(literal, BinaryLiteral):
                    left_predicate = literal.left_literal.get_predicate()
                    right_predicate = literal.right_literal.get_predicate()

                    if left_predicate in ["Bottom", "Top"]:
                        if right_predicate not in non_predicates:
                            continue # it is not recursive
                        for tmp_entity, tmp_context in ground_generator(literal.right_literal, dict(), D, D_index):
                            literal.set_entity([tmp_entity])
                            tmp_interval_list = apply(literal, D)
                            intervals = intervals + tmp_interval_list
                        contain_nrpredicate_flag = True 

                    elif right_predicate in ["Bottom", "Top"]:
                        if left_predicate not in non_predicates:
                            continue
                        for tmp_entity, tmp_context in ground_generator(literal.left_literal, dict(), D, D_index):
                            literal.set_entity([tmp_entity])
                            tmp_interval_list = apply(literal, D)
                            intervals = intervals + tmp_interval_list
                        contain_nrpredicate_flag = True 
                    else:
                        if left_predicate not in non_predicates or right_predicate not in non_predicates:
                            continue
                        for left_entity, tmp_context1 in ground_generator(literal.left_literal, dict(), D):
                            for right_entity, tmp_context2 in ground_generator(literal.right_literal, tmp_context1, D):
                                literal.set_entity([left_entity, right_entity])
                                tmp_interval_list = apply(literal, D)
                                intervals = intervals + tmp_interval_list
                        contain_nrpredicate_flag = True 
                else:
                    if literal.get_predicate() not in non_predicates:
                        continue
                    for tmp_entity, tmp_context in ground_generator(literal, dict(), D, D_index):
                        literal.set_entity(tmp_entity)
                        tmp_interval_list = apply(literal, D)
                        intervals = intervals + tmp_interval_list
                    contain_nrpredicate_flag = True 

                if len(intervals) != 0:
                    tmp_max_right = max([item.right_value for item in intervals])
                    min_right = min(tmp_max_right, min_right)
                
                elif len(intervals) == 0 and contain_nrpredicate_flag:
                    deleted_rules.add(rule)

            if min_right != float("inf"):
                observed_rules[min_right].append(rule)

    # backward propagation
    elif propagation == 2:
        for rule in rules:
            max_left = float("-inf")
            contain_nrpredicate_flag = False 
            for literal in rule.body:
                intervals = []
                if isinstance(literal, BinaryLiteral):
                    left_predicate = literal.left_literal.get_predicate()
                    right_predicate = literal.right_literal.get_predicate()

                    if left_predicate in ["Bottom", "Top"]:
                        if right_predicate not in non_predicates:
                            continue  # it is not recursive
                        for tmp_entity, tmp_context in ground_generator(literal.right_literal, dict(), D, D_index):
                            literal.set_entity([tmp_entity])
                            tmp_interval_list = apply(literal, D)
                            intervals = intervals + tmp_interval_list
                        contain_nrpredicate_flag = True 

                    elif right_predicate in ["Bottom", "Top"]:
                        if left_predicate not in non_predicates:
                            continue
                        for tmp_entity, tmp_context in ground_generator(literal.left_literal, dict(), D, D_index):
                            literal.set_entity([tmp_entity])
                            tmp_interval_list = apply(literal, D)
                            intervals = intervals + tmp_interval_list
                        contain_nrpredicate_flag = True 

                    else:
                        if left_predicate not in non_predicates or right_predicate not in non_predicates:
                            continue
                        for left_entity, tmp_context1 in ground_generator(literal.left_literal, dict(), D):
                            for right_entity, tmp_context2 in ground_generator(literal.right_literal, tmp_context1, D):
                                literal.set_entity([left_entity, right_entity])
                                tmp_interval_list = apply(literal, D)
                                intervals = intervals + tmp_interval_list
                        contain_nrpredicate_flag = True 
                else:
                    for tmp_entity, tmp_context in ground_generator(literal, dict(), D, D_index):
                        literal.set_entity([tmp_entity])
                        tmp_interval_list = apply(literal, D)
                        intervals = intervals + tmp_interval_list
                    contain_nrpredicate_flag = True 

                if len(intervals) != 0:
                    tmp_min_left = min([item.left_value for item in intervals])
                    min_right = max(tmp_min_left, max_left)
                
                elif len(intervals) == 0 and contain_nrpredicate_flag:
                   deleted_rules.add(rule)
                   break

            if max_left != float("-inf"):
                observed_rules[max_left].append(rule)

    observed_rules = sorted(observed_rules.items(), key=lambda item: item[0])
    for rule in deleted_rules:
        # delete the useless rule because one of its non-recursive body atom could not be grounded
        rules.remove(rule)

    return observed_rules




from meteor_reasoner.materialization.join_util import *
from meteor_reasoner.materialization.apply import *


def ground_generator(literal, context, D, D_index=None, delta_old=None, visited=False, flag=False):
    """
    This function is a generator, which generates entity and the context.
    Args:
        literal (a literal instance):
        context (a dictionary): store the mapping of  variables and constants
        D (a dictionary object): a global object storing all facts.

    Returns:
        entity and the context

    """
    predicate = literal.get_predicate()
    entity = copy.deepcopy(literal.get_entity())

    if visited:
        if predicate not in delta_old:
            return

        if len(entity) == 1 and entity[0].name == "nan":
            yield entity, dict()

        elif not contain_variable(entity):
            if predicate in delta_old and entity in delta_old[predicate]:
                yield entity, dict()

        elif not contain_variable_after_replace(entity, context):
            replaced_entity = []
            for term in entity:
                if term.type == "variable":
                    term.type = "constant"
                    term.name = context[term.name]
                    replaced_entity.append(term)
                else:
                    replaced_entity.append(term)

            if predicate in delta_old and tuple(replaced_entity) in delta_old[predicate]:
                yield tuple(replaced_entity), dict()
        else:
            for tmp_entity in delta_old[predicate]:
                tmp_context = dict()
                for term1, term2 in zip(entity, tmp_entity):
                    if term1.type == "constant" and term1.name != term2.name:
                        break
                    elif term1.type == "variable" and term1.name in context and context[term1.name] != term2.name:
                        break
                    elif term1.type == "variable" and term1.name in tmp_context and tmp_context[term1.name] != term2.name:
                        break
                    else:
                        if term1.type == "variable":
                           tmp_context[term1.name] = term2.name
                else:
                    yield tmp_entity, tmp_context

    elif not flag:
        if predicate not in D:
            return

        if len(entity) == 1 and entity[0].name == "nan":
            yield entity, dict()

        elif not contain_variable(entity):
            if predicate in D and entity in D[predicate]:
                yield entity, dict()

        elif not contain_variable_after_replace(entity, context):
            replaced_entity = []
            for term in entity:
                if term.type == "variable":
                    term.type = "constant"
                    term.name = context[term.name]
                    replaced_entity.append(term)
                else:
                    replaced_entity.append(term)
            replaced_entity = tuple(replaced_entity)
            if predicate in D and tuple(replaced_entity) in D[predicate]:
                yield tuple(replaced_entity), dict()
        else:
            if D_index is not None:
                index_str = []
                for i, term in enumerate(entity):
                    if term.type == "constant":
                        index_str.append(str(i) + "@" + term.name)
                    else:
                        if term.name in context:
                            index_str.append(str(i) + "@" + context[term.name])

                index_str = "||".join(index_str)
                if len(index_str) == 0:
                    for constant_entity in D[predicate]:
                        tmp_context = dict()
                        for term1, term2 in zip(entity, constant_entity):
                            if term1.type == "constant" and term1.name != term2.name:
                                break
                            elif term1.type == "variable" and term1.name in context and context[
                                term1.name] != term2.name:
                                break
                            elif term1.type == "variable" and term1.name in tmp_context and tmp_context[
                                term1.name] != term2.name:
                                break
                            else:
                                if term1.type == "variable":
                                    tmp_context[term1.name] = term2.name
                        else:
                            yield constant_entity, tmp_context
                else:
                    if index_str in D_index[predicate]:
                        for constant_entity in D_index[predicate][index_str]:
                            tmp_context = dict()
                            for term1, term2 in zip(entity, constant_entity):
                                if term1.type == "constant" and term1.name != term2.name:
                                    break
                                elif term1.type == "variable" and term1.name in context and context[
                                    term1.name] != term2.name:
                                    break
                                elif term1.type == "variable" and term1.name in tmp_context and tmp_context[
                                    term1.name] != term2.name:
                                    break
                                else:
                                    if term1.type == "variable":
                                        tmp_context[term1.name] = term2.name
                            else:
                                yield constant_entity, tmp_context

            else:
                for tmp_entity in D[predicate]:
                    tmp_context = dict()
                    for term1, term2 in zip(entity, tmp_entity):
                        if term1.type == "constant" and term1.name != term2.name:
                            break
                        elif term1.type == "variable" and term1.name in context and context[term1.name] != term2.name:
                            break
                        elif term1.type == "variable" and term1.name in tmp_context and tmp_context[
                            term1.name] != term2.name:
                            break
                        else:
                            if term1.type == "variable":
                                tmp_context[term1.name] = term2.name
                    else:
                        yield tmp_entity, tmp_context
    else:
        # index > visited
        if predicate not in D:
            return

        if len(entity) == 1 and entity[0].name == "nan":
            if predicate in delta_old and entity in delta_old[predicate]:
                if Interval.list_inclusion(D[predicate][entity], delta_old[predicate][entity]):
                    return
            yield entity, dict()

        elif not contain_variable(entity):
            if predicate in D and entity in D[predicate]:
                if predicate in delta_old and entity in delta_old[predicate]:
                    if Interval.list_inclusion(D[predicate][entity], delta_old[predicate][entity]):
                        return
                yield entity, dict()

        elif not contain_variable_after_replace(entity, context):
            replaced_entity = []
            for term in entity:
                if term.type == "variable":
                    term.type = "constant"
                    term.name = context[term.name]
                    replaced_entity.append(term)
                else:
                    replaced_entity.append(term)
            replaced_entity = tuple(replaced_entity)
            if predicate in D and tuple(replaced_entity) in D[predicate]:
                if predicate in delta_old and replaced_entity in delta_old[predicate]:
                    if Interval.list_inclusion(D[predicate][replaced_entity], delta_old[predicate][replaced_entity]):
                        return
                yield tuple(replaced_entity), dict()
        else:
            if D_index is not None:
                index_str = []
                for i, term in enumerate(entity):
                    if term.type == "constant":
                        index_str.append(str(i) + "@" + term.name)
                    else:
                        if term.name in context:
                            index_str.append(str(i) + "@" + context[term.name])

                index_str = "||".join(index_str)
                if len(index_str) == 0:
                    for constant_entity in D[predicate]:
                        tmp_context = dict()
                        for term1, term2 in zip(entity, constant_entity):
                            if term1.type == "constant" and term1.name != term2.name:
                                break
                            elif term1.type == "variable" and term1.name in context and context[
                                term1.name] != term2.name:
                                break
                            elif term1.type == "variable" and term1.name in tmp_context and tmp_context[
                                term1.name] != term2.name:
                                break
                            else:
                                if term1.type == "variable":
                                    tmp_context[term1.name] = term2.name
                        else:
                            if predicate in delta_old and constant_entity in delta_old[predicate]:
                                if Interval.list_inclusion(D[predicate][constant_entity],
                                                           delta_old[predicate][constant_entity]):
                                    continue
                            yield constant_entity, tmp_context
                else:
                    if index_str in D_index[predicate]:
                        for constant_entity in D_index[predicate][index_str]:
                            tmp_context = dict()
                            for term1, term2 in zip(entity, constant_entity):
                                if term1.type == "constant" and term1.name != term2.name:
                                    break
                                elif term1.type == "variable" and term1.name in context and context[
                                    term1.name] != term2.name:
                                    break
                                elif term1.type == "variable" and term1.name in tmp_context and tmp_context[
                                    term1.name] != term2.name:
                                    break
                                else:
                                    if term1.type == "variable":
                                        tmp_context[term1.name] = term2.name
                            else:
                                if predicate in delta_old and constant_entity in delta_old[predicate]:
                                    if Interval.list_inclusion(D[predicate][constant_entity],
                                                               delta_old[predicate][constant_entity]):
                                        continue
                                yield constant_entity, tmp_context

            else:
                for tmp_entity in D[predicate]:
                    tmp_context = dict()
                    for term1, term2 in zip(entity, tmp_entity):
                        if term1.type == "constant" and term1.name != term2.name:
                            break
                        elif term1.type == "variable" and term1.name in context and context[term1.name] != term2.name:
                            break
                        elif term1.type == "variable" and term1.name in tmp_context and tmp_context[
                            term1.name] != term2.name:
                            break
                        else:
                            if term1.type == "variable":
                                tmp_context[term1.name] = term2.name
                    else:
                        if predicate in delta_old and tmp_entity in delta_old[predicate]:
                            if Interval.list_inclusion(D[predicate][tmp_entity],
                                                       delta_old[predicate][tmp_entity]):
                                continue

                        yield tmp_entity, tmp_context


if __name__ == "__main__":
    from collections import defaultdict
    from meteor_reasoner.classes.term import Term
    from meteor_reasoner.materialization.index_build import build_index

    D = defaultdict(lambda: defaultdict(list))
    D["A"][tuple([Term("mike"), Term("nick")])] = [Interval(3, 4, False, False), Interval(6, 10, True, True)]
    D["B"][tuple([Term("nan")])] = [Interval(2, 8, False, False)]
    D_index = build_index(D)

    Delta = defaultdict(lambda: defaultdict(list))
    head = Atom("C", tuple([Term("nan")]))
    literal_a = Literal(Atom("A", tuple([Term("Y", "variable"), Term("X", "variable")])), [Operator("Boxminus", Interval(1, 2, False, False))])
    literal_b = Literal(Atom("B", tuple([Term("nan")])), [Operator("Diamondminus", Interval(0, 1, False, False))])

    body = [literal_a, literal_b]

    new_literal = BinaryLiteral(Atom("A", tuple([Term("X", "variable"), Term("Y", "variable")])), Atom("B", tuple([Term("nan")])), Operator("Since", Interval(1, 2, False, False)))
    body.append(new_literal)

    for atom_instance in ground_generator(literal_a, context={}, D=D, D_index=D_index):
        print(atom_instance)






from collections import deque
from meteor_reasoner.classes.atom import  Atom
from meteor_reasoner.classes.rule import *
from meteor_reasoner.classes.literal import *
import copy

prefix_ = "Build_"


def helper(i, operators, literal, build_dict, cnt, new_P):
    if i == -1:
        return literal

    else:
        if operators[i].name not in ["Diamondminus", "Diamondplus"]:
           body = copy.deepcopy(literal)
           body.operators = deque([operators[i]])

           if body in build_dict:
               tmp_head = build_dict[body]

           else:
               tmp_head = Atom(prefix_ + str(cnt[0]), literal.atom.entity)
               build_dict[body] = tmp_head
               cnt[0] = cnt[0] + 1

           if i == 0:
               return helper(i-1, operators, body)

           else:
               new_rule = Rule(tmp_head, [body])
               new_P.append(new_rule)

               literal.atom = tmp_head
               literal.operators = deque(list(literal.operators)[:i])
               return helper(i-1, operators, literal)

        else:

            body = copy.deepcopy(literal)
            body.operators = deque([operators[i]])

            if i == 0:
                top_atom = Atom("Top", ())
                left_literal = top_atom
                right_literal = body.atom

                if operators[i].name == "Diamondminus":
                    binary_literal = BinaryLiteral(left_literal, right_literal, Operator("Since", operators[i].interval))
                else:
                    binary_literal = BinaryLiteral(left_literal, right_literal, Operator("Until", operators[i].interval))

                return helper(i - 1, operators, binary_literal, build_dict, cnt, new_P)

            else:
                if body in build_dict:
                    tmp_head = build_dict[body]

                else:
                    tmp_head = Atom(prefix_ + str(cnt[0]), literal.atom.entity)
                    build_dict[body] = tmp_head
                    cnt[0] = cnt[0] + 1

                top_atom = Atom("Top", ())
                left_literal = top_atom
                right_literal = body.atom

                if operators[i].name == "Diamondminus":
                    binary_literal = BinaryLiteral(left_literal, right_literal, Operator("Since", operators[i].interval))
                else:
                    binary_literal = BinaryLiteral(left_literal, right_literal, Operator("Until", operators[i].interval))

                new_rule = Rule(tmp_head, [binary_literal])
                new_P.append(new_rule)
                literal.atom = tmp_head
                literal.operators = deque(list(literal.operators)[:i])
                return helper(i - 1, operators, literal, build_dict, cnt)


def normalize(P):
    """
    This function normalize each rule in P into four normal forms, Since, Until, Boxminus, Boxplus.
    Args:
        P (list of Rule instances):

    Returns:
        list of Rule instances

    """
    cnt = [0]
    build_dict = {}
    new_P = []

    for rule in P:
        head = rule.head
        body = rule.body
        new_body = []
        for literal in body:

            if isinstance(literal, BinaryLiteral):
                new_body.append(literal)
            elif isinstance(literal, Atom):
                new_body.append(literal)
            else:
                normal_flag = False
                for operator in literal.operators:
                    if operator.name in ["Diamondminus", "Diamondplus"]:
                        normal_flag = True
                        break
                if not normal_flag:
                    new_body.append(literal)

                else:
                    operators = literal.operators
                    new_literal = helper(len(operators)-1, operators, literal, build_dict, cnt, new_P)
                    new_body.append(new_literal)

        rule = Rule(head, new_body)
        new_P.append(rule)

    return new_P








from meteor_reasoner.materialization.apply import *
from meteor_reasoner.materialization.index_build import *
from meteor_reasoner.utils.parser import *


def test_apply():
    D = defaultdict(lambda: defaultdict(list))
    D["A"][tuple([Term("mike")])] = [Interval(3, 4, False, False), Interval(6, 10, True, True)]
    D["C"][tuple([Term("mike"), Term("jack")])] = [Interval(3, 4, False, False), Interval(6, 10, True, True)]
    D["B"][tuple([Term("nan")])] = [Interval(2, 8, False, False)]
    D_index = defaultdict(lambda: defaultdict(list))
    build_index(D, D_index)

    raw_literal = "BSince[2,3]Diamondminus[1,2]Boxminus[1,2]A(mike)"
    literal = parse_literal(raw_literal)
    print(str(literal))

    T = apply(literal, D)
    print([str(interval) for interval in T])






from meteor_reasoner.utils.conversion import *
from multiprocessing.pool import ThreadPool
from meteor_reasoner.automata.buichi_automata import *
from meteor_reasoner.automata.utils import *
from meteor_reasoner.utils.normalize import *
from meteor_reasoner.materialization.index_build import *
from meteor_reasoner.materialization.t_operator import naive_immediate_consequence_operator
from meteor_reasoner.utils.operate_dataset import print_dataset


def contain_new_in_limit(D, delta_new, delta_old, D_index, limit):
    for head_predicate in delta_new:
        for entity, T in delta_new[head_predicate].items():
            if head_predicate not in D or entity not in D[head_predicate]:
                D[head_predicate][entity] = D[head_predicate][entity] + T
                delta_old[head_predicate][entity] = delta_old[head_predicate][entity] + T
                # update index
                for i, item in enumerate(entity):
                    D_index[head_predicate][str(i) + "@" + item.name].append(entity)
                if len(entity) > 2:
                    for i, item1 in enumerate(entity):
                        for j, item2 in enumerate(entity):
                            if j <= i:
                                continue
                            D_index[head_predicate][str(i) + "@" + item1.name + "||" + str(j) + "@" + item2.name].append(entity)
            else:
                coalesced_T = coalescing(T + D[head_predicate][entity])
                if not Interval.compare(coalesced_T, D[head_predicate][entity]):
                    for interval1 in coalesced_T:
                        flag = True
                        for interval2 in D[head_predicate][entity]:
                            if interval1.left_value == interval2.left_value and interval1 == interval2:
                                flag = False
                                break
                            elif interval2.left_value > interval1.left_value:
                                break
                        if flag:
                            delta_old[head_predicate][entity].append(interval1)
                    D[head_predicate][entity] = coalesced_T
    fixpoint = True
    if len(delta_old) != 0:
        fixpoint = False
        coalescing_d(D)
        coalescing_d(delta_old)
        for predicate in delta_old:
            for entity in delta_old[predicate]:
                for interval in delta_old[predicate][entity]:
                    if not Interval.inclusion(interval, limit):
                        return fixpoint, False
    return fixpoint, True


def automata(D, program, timeout=120):
    pool1 = ThreadPool(processes=1)
    async_result1 = pool1.apply_async(consistency, args=(D, program))

    try:
        result = async_result1.get(timeout=timeout)  # setting 120 as the timeout time
        return result
    except:
        return "timeout"


def consistency(D, program, F=None):
    contain_bottom = False
    for rule in program:
        # if the program contains no Bottom in the head, it is definitely consistent
        if rule.head.get_predicate() == "Bottom":
            contain_bottom = True
            break
    if F is None and not contain_bottom:
        return True

    if F is not None:
        # construct a new fact and a new rule, which will be added to the daset and the program, respectively.
        # in this way, we change the fact entailment problem to the consistency checking problem
        new_rule, new_atom, _ = euqal_conversion(F)
        program.append(new_rule)
        D[new_atom.predicate][new_atom.entity] = [new_atom.interval]
        #program = normalize(program)
    
    # must_literals store those immediate results calculated during the materialisation process
    # in this way, we can know the time range for which some metric atoms hold.
    must_literals = defaultdict(list)
    coalescing_d(D)
    D_index = build_index(D)
    relevant_rules = program
    CF = CycleFinder(program=program)
    non_recursive_predicates = CF.get_non_recursive_predicates()
    involved_predicates = [predicate for predicate in CF.predicate2id if predicate not in ["Bottom", "Top"]]
    automata_predicates = list(set(involved_predicates) - set(non_recursive_predicates))
    d_atoms, d_unbounded_literals = extract_dataset(D, involved_predicates, automata_predicates)
    
    # print("bounded atoms:")
    # for atom in d_atoms:
    #     print(atom)
    # print("unbounded atoms:")
    # for atom in d_unbounded_literals:
    #     print(atom)
    # exit()

    points, x = get_dataset_points_x(D)
    constants = get_constants(D, program)
    
    # print("points:", points)
    # print("x:", x)
    # for constant in constants:
    #     print(constant)
    # exit()

    z, gcd = get_gcd(relevant_rules)
    left_dict, right_dict = construct_left_right_pattern(points, gcd)

    while True:
        delta_new = naive_immediate_consequence_operator(program, D, D_index, must_literals=must_literals)
        delta_old = defaultdict(lambda: defaultdict(list))
        fixpoint, flag = contain_new_in_limit(D, delta_new, delta_old, D_index, limit=Interval(-4 * x - 4 * z,  4 * x + 4 * z, False, False))
        if not flag:
            break

    if "Bottom" in D:
        return False

    if fixpoint:
        return True

    remove_redundant_atoms = []
    for atom, intervals in must_literals.items():
        if isinstance(atom, Literal) and len(atom.operators) == 0:
            if atom.atom in must_literals:
                must_literals[atom.atom] += must_literals[atom]
                remove_redundant_atoms.append(atom)
    for atom in remove_redundant_atoms:
        del must_literals[atom]

    for key, value in must_literals.items():
        value = coalescing(value)
        must_literals[key] = value
    

    p_literals = []
    for rule in relevant_rules:
        for literal in rule.body:
            if isinstance(literal, BinaryLiteral) or (isinstance(literal, Literal) and len(literal.operators) != 0):
                p_literals.append(literal)
    
    automata = BuchiAutomata(D=D, program=relevant_rules,
                             p_literals=p_literals,
                             unbounded_literals=d_unbounded_literals,
                             involved_predicates=involved_predicates,
                             automata_predicates=automata_predicates,
                             left_dict=left_dict,
                             right_dict=right_dict,
                             constants=constants,
                             x=x, z=z, gcd=gcd, points=points)

    automata.build_prior(must_literals=must_literals)
    flag = automata.consistency_check()
    return flag


if __name__ == "__main__":
    #raw_data = "B(a,b)@[2,30]\nB(c,b)@[4,20]\nA(a)@[10,20]\nD(b)@[13,24]\nD(a)@[3,5]".split("\n")
    # raw_data = ["Backup@0"]
    # D = load_dataset(raw_data)
    # D_index = defaultdict(lambda: defaultdict(list))
    # build_index(D, D_index)
    # #raw_program ="A(X):-SOMETIME[1,2]B(X,Y)\n C(X):-ALWAYS[-4,0]A(X)\nC(X):-SOMETIME[-2,-1]D(X)\nD(X):-SOMETIME[-2,-1]C(X)".split("\n")
    # raw_program = ["Backup :- SOMETIME[-1]Backup"]
    # Program = load_program(raw_program)
    # #fact = "D(a)@[-10,3]"
    # fact = "Backup@-1"
    # fact = parse_str_fact(fact)
    # F = Atom(fact[0], fact[1], fact[2])
    #
    # flag = consistency(D, Program, F)
    # print("flag:", flag)


    #raw_data = ["P@0", "R@0"]
    raw_data = ["P@0","Q@0"]
    raw_data = ["Alive(adam)@0"]
    #raw_data = ["Alive(adam)@0"]
    D = load_dataset(raw_data)
    D_index = defaultdict(lambda: defaultdict(list))
    build_index(D, D_index)
    # raw_program ="A(X):-SOMETIME[1,2]B(X,Y)\n C(X):-ALWAYS[-4,0]A(X)\nC(X):-SOMETIME[-2,-1]D(X)\nD(X):-SOMETIME[-2,-1]C(X)".split("\n")
    #raw_program = ["P :- SOMETIME[-1,0] P", "R :- SOMETIME[-1,0] R", "Q :- ALWAYS[0,+inf) P", "Q :- ALWAYS[0,+inf) R"]
    raw_program = ["ALWAYS[0,1]P:- P", "R:- ALWAYS[0,+inf)P, Q","ALWAYS[0,1]R:- R", "S:-ALWAYS[0,+inf)R"]
    raw_program = ["ALWAYS[0,1]Alive(X):-Alive(X)", "Immortal(X):-ALWAYS[0,+inf)Alive(X)"]
    #raw_program = ["ALWAYS[0,1] Alive(X) :-  Alive(X)", "Immortal(X) :- ALWAYS[0,+inf) Alive(X)"]
    Program = load_program(raw_program)
    # fact = "D(a)@[-10,3]"
    #fact = "Immortal(adam)@-1"
    fact = "S@3"
    fact = "Immortal(adam)@0"
    fact = parse_str_fact(fact)
    F = Atom(fact[0], fact[1], fact[2])

    flag = consistency(D, Program, F)
    print("flag:", flag)






from meteor_reasoner.materialization.t_operator import *
from meteor_reasoner.materialization.coalesce import *
from meteor_reasoner.materialization.index_build import *
from meteor_reasoner.utils.propagation import check_propagation
from meteor_reasoner.graphutil.temporal_dependency_graph import CycleFinder
import time
from meteor_reasoner.materialization.utils import no_new_facts, pre_calculate_threshold, entail_same_nonrecursive_predicates
from meteor_reasoner.utils.operate_dataset import save_dataset_to_file


def calculate_redundancy(delta, old):
    cnt = 0
    for predicate in delta:
        for entity in delta[predicate]:
            if predicate not in old or entity not in old[predicate]:  # all fact w.r.t the predicate are redundant
                cnt += len(delta[predicate][entity])
            else:
                for delta_interval in delta[predicate][entity]:
                    for old_interval in old[predicate][entity]:
                        if Interval.intersection(delta_interval, old_interval) is not None:
                            break
                    else:
                        cnt += 1
    return cnt


def seminaive_combine(D, delta_new, delta_old, D_index=None):
    for head_predicate in delta_new:
        for entity, T in delta_new[head_predicate].items():
            if head_predicate not in D or entity not in D[head_predicate]:
                D[head_predicate][entity] = D[head_predicate][entity] + T
                delta_old[head_predicate][entity] = delta_old[head_predicate][entity] + T
                # update index
                for i, item in enumerate(entity):
                    D_index[head_predicate][str(i) + "@" + item.name].append(entity)
                if len(entity) > 2:
                    for i, item1 in enumerate(entity):
                        for j, item2 in enumerate(entity):
                            if j <= i:
                                continue
                            D_index[head_predicate][str(i) + "@" + item1.name + "||" + str(j) + "@" + item2.name].append(entity)
            else:
                coalesced_T = coalescing(T + D[head_predicate][entity])
                if not Interval.compare(coalesced_T, D[head_predicate][entity]):
                    for interval1 in coalesced_T:
                        flag = True
                        for interval2 in D[head_predicate][entity]:
                            if interval1.left_value == interval2.left_value and interval1 == interval2:
                                flag = False
                                break
                            elif interval2.left_value > interval1.left_value:
                                break
                        if flag:
                            delta_old[head_predicate][entity].append(interval1)
                    D[head_predicate][entity] = coalesced_T
    fixpoint = True
    if len(delta_old) != 0:
        coalescing_d(D)
        coalescing_d(delta_old)
        fixpoint = False
    return fixpoint


def naive_combine(D, delta_new, D_index=None):
    fixpoint = True
    for head_predicate in delta_new:
        for entity, T in delta_new[head_predicate].items():
            if head_predicate not in D or entity not in D[head_predicate]:
                fixpoint = False
                try:
                    D[head_predicate][entity] = D[head_predicate][entity] + T
                except:
                    print("debug")
                    D[head_predicate][entity] = D[head_predicate][entity] + T
                # update index
                for i, item in enumerate(entity):
                    D_index[head_predicate][str(i) + "@" + item.name].append(entity)
                if len(entity) > 2:
                    for i, item1 in enumerate(entity):
                        for j, item2 in enumerate(entity):
                            if j <= i:
                                continue
                            D_index[head_predicate][str(i) + "@" + item1.name + "||" + str(j) + "@" + item2.name].append(entity)
            else:
                coalesced_T = coalescing(T + D[head_predicate][entity])
                if fixpoint:
                    if coalesced_T != D[head_predicate][entity]:
                        fixpoint = False
                D[head_predicate][entity] = coalesced_T

    if not fixpoint:
         coalescing_d(D)

    return fixpoint


def materialize(D, rules, mode="seminaive", K=100, logger=None, must_literals=None, metrics=None):
    """
    The function implements the materialization operation.
    Args:
        D (dictionary object):
        rules (list of Rule instances):
        K (int): the nunber of steps for exectuting the materialization. By default=100.
    Returns:
        The
    """
    D_index = build_index(D)
    delta_old = D
    if mode == "opt":
        return opt_materialize(D, rules, delta_old=delta_old, D_index=D_index, K=K)
    elif mode == "seminaive":
        seminaive = True
    else:
        seminaive = False

    k = 0
    if logger is not None:
        start_time = time.time()
        calc_time = 0.0

    while k < K:
        print("Iteration:", k)
        k += 1
        if seminaive:
            delta_new = seminaive_immediate_consequence_operator(rules, D, D_index, delta_old=delta_old, must_literals=must_literals)
        else:
            delta_new = naive_immediate_consequence_operator(rules, D, D_index, must_literals=must_literals)

        if seminaive:
            delta_old = defaultdict(lambda: defaultdict(list))
            fixpoint = seminaive_combine(D, delta_new, delta_old, D_index)
            if logger is not None:
                tmp_start_time = time.time()
                coalescing_d(delta_new)
                number_of_redundant_facts = calculate_redundancy(delta_new, delta_old)
                total_number = 0
                for predicate in D:
                    for entity in D[predicate]:
                        total_number += len(D[predicate][entity])
                calc_time += time.time() - tmp_start_time
                logger.info("Iteration={}, t={}, D={}, n={}".format(k, time.time() - start_time - calc_time, total_number, number_of_redundant_facts))
        else:
            if logger is not None:
                delta_old = defaultdict(lambda: defaultdict(list))
                fixpoint = seminaive_combine(D, delta_new, delta_old, D_index)
                coalescing_d(delta_new)
                number_of_redundant_facts = calculate_redundancy(delta_new, delta_old)
                total_number = 0
                for predicate in D:
                    for entity in D[predicate]:
                        total_number += len(D[predicate][entity])
                logger.info("Iteration={}, t={}, D={}, n={}".format(k, time.time() - start_time - calc_time, total_number, number_of_redundant_facts))
            else:
                fixpoint = naive_combine(D, delta_new, D_index)

        if fixpoint:
            return True

    return False


def opt_materialize(D, rules, D_index, delta_old, K=1, logger=None):
    CF = CycleFinder(program=rules)
    non_predicates = CF.get_non_recursive_predicates()
    nr_program = list()
    r_program = list()
    for rule in rules:
        for body_atom in rule.body:
            if isinstance(body_atom, BinaryLiteral):
                left_atom = body_atom.left_literal
                right_atom = body_atom.right_literal
                if left_atom.get_predicate() not in non_predicates or right_atom.get_predicate() not in non_predicates:
                    r_program.append(rule)
                    break  
            else:
                if body_atom.get_predicate() not in non_predicates:
                    r_program.append(rule)
                    break 
        else:
             nr_program.append(rule)
    flag = 0
    k = 0
    if logger is not None:
        start_time = time.time()
        calc_time = 0.0

    while k < K:
        print("Iteration:", k)
        k += 1
        delta_new = seminaive_immediate_consequence_operator(rules, D, D_index, delta_old=delta_old)
        if flag == 0:
            if entail_same_nonrecursive_predicates(D, delta_new, non_predicates):
                flag = 1
                for rule in nr_program:
                    rules.remove(rule)

                if len(r_program) != 0:
                    propagation = check_propagation(r_program)
                    if propagation == 1 or propagation == 2:
                        observed_rules = pre_calculate_threshold(rules, propagation, D, D_index, non_predicates)
                else:
                    propagation = 0

        elif flag == 1 and propagation == 1:
            if len(observed_rules) != 0:
                for i, item in enumerate(observed_rules):
                    if no_new_facts(delta_new, D, Interval(float("-inf"), item[0], True, False)):
                        for rule in item[1]:
                            try:
                                rules.remove(rule)
                            except:
                                continue
                    else:
                        observed_rules = observed_rules[i:]
                        break
                else:
                    observed_rules = observed_rules[i + 1:]

        elif flag == 1 and propagation == 2:
            if len(observed_rules) != 0:
                for i, item in enumerate(observed_rules):
                    if no_new_facts(delta_new, D, Interval(item[0], float("inf"), False, True)):
                        for rule in item[1]:
                             rules.remove(rule)
                    else:
                        observed_rules = observed_rules[i:]
                        break
                else:
                    observed_rules = observed_rules[i + 1:]
        #print("len of rules:", len(rules))
        # if len(rules) < 15:
        #    for rule in rules:
        #         print(rule)
        delta_old = defaultdict(lambda: defaultdict(list))
        fixpoint = seminaive_combine(D, delta_new, delta_old, D_index)
        if logger is not None:
            tmp_start_time = time.time()
            coalescing_d(delta_new)
            number_of_redundant_facts = calculate_redundancy(delta_new, delta_old)
            total_number = 0
            for predicate in D:
                for entity in D[predicate]:
                    total_number += len(D[predicate][entity])
            calc_time += time.time() - tmp_start_time
            logger.info("Iteration={}, t={}, D={}, n={}".format(k, time.time() - start_time - calc_time,
                                                                total_number, number_of_redundant_facts))
        if fixpoint:
            save_dataset_to_file("recursive_dataset.txt", D)
            return True

    return False


if __name__ == "__main__":
    from meteor_reasoner.utils.operate_dataset import print_dataset
    from meteor_reasoner.utils.loader import *

    # start_time = time.time()
    # raw_program = [
    #     "P(X,Y):-Diamondminus[0,1]P(X,Y)",
    #     "R(Y):-Boxminus[1,2]Q(Y),H(Z),P(X,Y)",
    #     "Q(X):-K(X,Y)"
    # ]
    # raw_data = [
    #     "P(a,b)@0",
    #     "K(b,c)@[1,2]",
    #     "H(c)@[2,5]"
    # ]
    # D = load_dataset(raw_data)
    # coalescing_d(D)
    # D_index = build_index(D)
    # program = load_program(raw_program)
    # opt_materialize(D, rules=program, D_index=D_index, delta_old=D, K=1000)
    # print("Optimized method time: ", time.time()-start_time)
    #
    # exit()
    start_time = time.time()
    raw_program = [
        "A(X):-Diamondminus[1,1]A(X)",
        "B(Y):-Diamondminus[1,1]B(Y)",
        "C(X):-A(X), B(X), D(X)"
    ]
    raw_data = [
        "A(a)@1",
        "B(a)@1",
        "D(a)@(-inf,+inf)"
    ]

    D = load_dataset(raw_data)
    coalescing_d(D)
    D_index = build_index(D)
    program = load_program(raw_program)

    materialize(D, rules=program, D_index=D_index, delta_old=D, seminaive=True, K=3)
    print_dataset(D)
    print("Seminaive method time: ", time.time() - start_time)

    start_time = time.time()
    raw_program = [
        "A(X):-Diamondminus[1,1]A(X)",
        "B(Y):-Diamondminus[1,1]B(Y)",
        "C(X):-A(X), B(X), D(X)"
    ]
    raw_data = [
        "A(a)@1",
        "B(a)@1",
        "D(a)@(-inf,+inf)"
    ]

    D = load_dataset(raw_data)
    coalescing_d(D)
    D_index = build_index(D)
    program = load_program(raw_program)
    materialize(D, rules=program, D_index=D_index, delta_old=D, seminaive=False, K=3)
    print("Naive method time: ", time.time() - start_time)
    print_dataset(D)
    exit()

    start_time = time.time()
    raw_program = [
        "P(X,Y):-Diamondminus[0,1]P(X,Y)",
        "R(Y):-Q(Y),H(Z),P(X,Y)",
        "Q(X):-K(X,Y)"
    ]
    raw_data = [
        "P(a,b)@0",
        "K(b,c)@[1,2]",
        "H(c)@[2,5]"
    ]
    D = load_dataset(raw_data)
    coalescing_d(D)
    D_index = build_index(D)
    program = load_program(raw_program)
    materialize(D, rules=program, D_index=D_index, K=1000)
    print("naive method time: ", time.time() - start_time)


    exit()
    raw_data = ["P(mike,mike)@1"]
    raw_program = ["P(X,X):- Diamondminus[1,1]P(X,X)"]
    D = load_dataset(raw_data)
    coalescing_d(D)
    D_index = build_index(D)
    Program = load_program(raw_program)

    start_time = time.time()
    materialize(D, Program, D_index, K=100)
    print_dataset(D)
    end_time = time.time()
    print("The naive mat time:", end_time-start_time)

    raw_data = ["P(mike,mike)@1"]
    raw_program = ["P(X,X):- Diamondminus[1,1]P(X,X)"]
    D = load_dataset(raw_data)
    coalescing_d(D)
    D_index = build_index(D)
    Program = load_program(raw_program)
    start_time = time.time()
    materialize(D, Program, D_index, seminaive=True, delta_old=D, K=100)
    print_dataset(D)
    end_time = time.time()
    print("The seminaive mat time:", end_time - start_time)



















from meteor_reasoner.materialization.index_build import build_index
from meteor_reasoner.utils.loader import *
from meteor_reasoner.utils.operate_dataset import get_min_max_rational
from meteor_reasoner.materialization.coalesce import coalescing_d, coalescing
from meteor_reasoner.materialization.t_operator import naive_immediate_consequence_operator
import time
import pickle
import sys
import decimal
import argparse
from meteor_reasoner.utils.operate_dataset import print_dataset
from meteor_reasoner.stream.utils import *
from meteor_reasoner.stream.stream_generator import Stream_Generator
import logging
logger = logging.getLogger()
logger.setLevel(logging.INFO)

output_file_handler = logging.FileHandler("./log/mock.log")
stdout_handler = logging.StreamHandler(sys.stdout)
logger.addHandler(output_file_handler)
logger.addHandler(stdout_handler)

parser = argparse.ArgumentParser()
parser.add_argument("--datapath", default="../datasets/jsw/onlane.txt",  type=str, help="Input the dataset path (json format)")
parser.add_argument("--rulepath", default="./programs/traffic_shortbreak.txt", type=str, help="Input the program path")
parser.add_argument("--datasetname", default="traffic_shortbreak", type=str, help="Input the save path")
parser.add_argument("--target", default="ShortStop", type=str, help="Input the program path")
args = parser.parse_args()

target_predicate = args.target
with open(args.rulepath) as file:
    raw_program = file.readlines()
    program = load_program(raw_program)
    predicates = set()
    for rule in program:
        predicates.add(rule.head.get_predicate())
        for literal in rule.body:
            if isinstance(literal, BinaryLiteral):
                predicates.add(literal.left_literal.get_predicate())
                predicates.add(literal.right_literal.get_predicate())
            else:
                predicates.add(literal.get_predicate())


start_time = time.time()
with open(args.datapath) as file:
    raw_data = []
    for line in file:
        for predicate in predicates:
            if line.startswith(predicate):
                 raw_data.append(line)
                 break
    print("Lines of facts:", len(raw_data))
    D = load_dataset(raw_data)
    for predicate in D:
        for entity in D[predicate]:
            interval_list = []
            for item in D[predicate][entity]:
                value = (item.left_value + item.right_value) * 10 / 2
                interval_list.append(Interval(value, value, False, False))
            D[predicate][entity] = interval_list
    coalescing_d(D)
    D_index = build_index(D)

min_val, max_val = get_min_max_rational(D)
max_val = 2000
print("Min_value={}, Max_value={} in the given dataset".format(min_val, max_val))

time_points = []
point = min_val - decimal.Decimal(10)
while point <= max_val:
    point += decimal.Decimal(10)
    time_points.append(point)

print("Loading ends!={}s".format(time.time()-start_time))
t_program = decimal.Decimal(get_maximum_rational_number(program))
step = 1

coalescing_d(D)

t = -1
window = [decimal.Decimal(-1 - t_program), decimal.Decimal(-1)]
window_facts = defaultdict(lambda: defaultdict(list))

print("============Stream Reasoning Starts============\n")
start_time = time.time()
run_times = []
window_size = []
window_size_raw = []
in_streams_num = []
out_streams_num = []
rule_fired = []
apply_num = []

stream_generator = Stream_Generator(time_points, Data=D)

for streams, t_next in stream_generator.generator():
    # keep the window unchange
    if len(streams) == 0:
        continue
    logger.info("Streams in at the time point {}".format(t_next))
    logger.info("Number of streams: {}".format(len(streams)))

    in_streams_num.append(len(streams))
    # calculate the number of facts in the window
    # end of the calculation
    add_streams(window_facts, streams, t_next) # add the stream to window_facts

    window_index = build_index(window_facts)
    new_area = Interval(window[1], t_next, True, False)

    stream_out_cnt = 0
    iteration_step = 0

    raw_point_size = 0
    point_size = 0
    while True:
        #print_dataset(window_facts)
        print("Iteration_step:", iteration_step)
        iteration_step += 1
        one_start_time = time.time()
        delta_new = naive_immediate_consequence_operator(D=window_facts, rules=program, D_index=window_index)

        # stream_out
        streams_out = trim_delta(window_facts, delta_new, new_area)
        one_time = time.time() -one_start_time
        if len(streams_out) == 0:
            # trim to the window size
            window = [t_next-t_program, t_next]
            next_window = Interval(t_next-t_program, t_next, False, False)
            window_facts = trim_window(window_facts, next_window)
            logger.info("Streams out at the time point {}".format(t_next))
            logger.info("Number of out streams: {}".format(stream_out_cnt))
            out_streams_num.append(stream_out_cnt)
            logger.info("Run time is:{}".format(time.time() - start_time-one_time))
            run_times.append(time.time()-start_time)
            break
        for point in time_points:
            if target_predicate in streams_out:
                for entity in streams_out[target_predicate]:
                    for interval in streams_out[target_predicate][entity]:
                        if interval.left_value < point < interval.right_value or (interval.left_value == point
                                                                                  and not interval.left_open) or\
                                (interval.right_value == point and not interval.right_open):
                                stream_out_cnt += 1
                                break

        # === before coalescing =====
        raw_fact_num = 0
        for predicate in window_facts:
            for entity in window_facts[predicate]:
                raw_fact_num += len(window_facts[predicate][entity])

        for predicate in streams_out:
            for entity in streams_out[predicate]:
                raw_fact_num += len(streams_out[predicate][entity])

        raw_point_size = max(raw_fact_num, raw_point_size)

        # === before coalescing ========

        merge_streams(window_facts, streams_out)
        window_index = build_index(window_facts)
        fact_num = 0
        for predicate in window_facts:
            for entity in window_facts[predicate]:
                fact_num += len(window_facts[predicate][entity])
        point_size = min(fact_num, point_size) if point_size != 0 else fact_num

    window_size_raw.append(raw_point_size)
    window_size.append(point_size)
    logger.info("Number of window facts: uncoalesced={}, coalesced={}".format(raw_point_size, point_size))
    print("Iteration_step:", iteration_step)


results = {
    "run_times": run_times,
    "window_size": window_size,
    "window_size_raw": window_size_raw,
    "stream_in": in_streams_num,
    "stream_out": out_streams_num,
    "min_value": min_val,
    "max_value": max_val
}

logger.info(results)
from datetime import datetime
now = datetime.now()

output_path = "./results/sr_" + args.datasetname + ".pkl"
pickle.dump(results, open(output_path , "wb"))
print("============Stream Reasoning Ends============\n")






from meteor_reasoner.materialization.index_build import *
from meteor_reasoner.materialization.coalesce import *
from meteor_reasoner.materialization.t_operator import naive_immediate_consequence_operator
from meteor_reasoner.utils.ruler_interval import *
from meteor_reasoner.canonical.class_common_fragment import CommonFragment


def find_common_fragment(D1, D2, rules, varrho):
    points, min_x, max_x = get_dataset_points_x(D2, min_x_flag=True)
    _, gcd = get_gcd(rules)
    _, initial_window_ruler_intervals = get_initial_ruler_intervals(points, left_border=min_x, right_border=max_x, gcd=gcd)
    left_point = Interval(varrho.left_value, varrho.left_value, False, False)
    right_point = Interval(varrho.right_value, varrho.right_value, False, False)
    left_i = initial_window_ruler_intervals.index(left_point)
    right_i = initial_window_ruler_intervals.index(right_point)

    middle_i = left_i
    while middle_i <= right_i:
        flag = False
        ruler_interval = initial_window_ruler_intervals[middle_i]
        for predicate in D2:
            for entity in D2[predicate]:
                if interval_intesection_intervallist(ruler_interval, D2[predicate][entity]):
                    if predicate not in D1 or entity not in D1[predicate] or not \
                            interval_intesection_intervallist(ruler_interval, D1[predicate][entity]):
                        flag = True
                        break
            if flag:
                return None, None
        middle_i += 1

    left_border = None
    while left_i > 0:
        left_i -= 1
        ruler_interval = initial_window_ruler_intervals[left_i]
        for predicate in D2:
            flag = False
            for entity in D2[predicate]:
                if interval_intesection_intervallist(ruler_interval, D2[predicate][entity]):
                    if predicate not in D1 or entity not in D1[predicate] or not \
                            interval_intesection_intervallist(ruler_interval, D1[predicate][entity]):
                        left_border = ruler_interval
                        flag = True
                        break
            if flag:
                break

    right_border = None
    while right_i < len(initial_window_ruler_intervals)-1:
        right_i += 1
        ruler_interval = initial_window_ruler_intervals[right_i]
        flag = False
        for predicate in D2:
            for entity in D2[predicate]:
                if interval_inclusion_intervallist(ruler_interval, D2[predicate][entity]):
                    if predicate in D1 and entity in D1[predicate] and interval_inclusion_intervallist(
                            ruler_interval, D1[predicate][entity]):
                        continue
                    else:
                        if ruler_interval.left_open:
                            right_border = Interval(ruler_interval.left_value, ruler_interval.left_value, False, False)
                        else:
                            right_border = Interval(ruler_interval.left_value, ruler_interval.left_value, True, True)
                        flag = True
                        break
            if flag:
                break
        if flag:
            break

    if left_border is None:
        left_value, left_open = decimal.Decimal("-inf"), True
    else:
        left_value, left_open = left_border.left_value, left_border.left_open

    if right_border is None:
        right_value, right_open = decimal.Decimal("inf"), True
    else:
        right_value, right_open = right_border.left_value, right_border.left_open

    varrho_left_range = None
    if Interval.is_valid_interval(left_value, varrho.left_value, left_open, True):
        varrho_left_range = Interval(left_value, varrho.left_value, left_open, True)
    varrho_right_range = None
    if Interval.is_valid_interval(varrho.right_value, right_value, True, right_open):
        varrho_right_range = Interval(varrho.right_value, right_value, True, right_open)

    return varrho_left_range, varrho_right_range


def has_same_pattern(ruler_intervals1, ruler_intervals2):
    """
    Check whether the give two ruler interval lists have the same pattern
    Args:
        ruler_intervals1: a list of ruler-intervals
        ruler_intervals2: a list of ruler-intervals

    Returns:
        True or False
    """
    if len(ruler_intervals1) != len(ruler_intervals2):
        return False
    if ruler_intervals1[0].left_open != ruler_intervals2[0].left_open or \
            ruler_intervals1[-1].right_open != ruler_intervals2[-1].right_open:
        return False
    for ruler1, ruler2 in zip(ruler_intervals1, ruler_intervals2):
        if abs(ruler1.right_value-ruler1.left_value) != abs(ruler2.right_value - ruler2.left_value):
            return False
    return True


def has_same_facts(ruler_intervals1, ruler_intervals2, D):
    """
    Check whether the two same-pattern ruler lists have the same facts at each corresponding ruler-interval
    Args:
        ruler_intervals1: a list of ruler-intervals
        ruler_intervals2: a list of ruler-intervals
        D: contain all relational facts
    Returns:
        True or False
    """
    for ruler1, ruler2 in zip(ruler_intervals1, ruler_intervals2):
        for predicate in D:
            for entity in D[predicate]:
                if (interval_inclusion_intervallist(ruler1, D[predicate][entity]) and \
                        not interval_inclusion_intervallist(ruler2, D[predicate][entity])) or \
                        (interval_inclusion_intervallist(ruler2, D[predicate][entity]) and
                         not interval_inclusion_intervallist(ruler1, D[predicate][entity])):
                    return False
    return True


def build_left_ruler_intervals(left_interval_range, CR):
    big_ruler_intervals = CR.left_initial_ruler_intervals[:]
    base_index = big_ruler_intervals.index(Interval(CR.min_x, CR.min_x, False, False))
    starting_ruler_interval = big_ruler_intervals[base_index-2]
    while big_ruler_intervals[0].left_value >= left_interval_range.left_value:
        ruler_interval_len = []
        cnt = 0
        if big_ruler_intervals[0].left_open:
            new_ruler_interval = Interval(big_ruler_intervals[0].left_value, big_ruler_intervals[0].left_value,
                                          False, False)
            if new_ruler_interval.left_value > left_interval_range.left_value:
               big_ruler_intervals = [new_ruler_interval] + big_ruler_intervals
               continue
            elif new_ruler_interval.left_value == left_interval_range.left_value and new_ruler_interval.left_open == left_interval_range.left_open:
                big_ruler_intervals = [new_ruler_interval] + big_ruler_intervals
                continue
            else:
                break

        for i in range(len(big_ruler_intervals)):
            if big_ruler_intervals[i].left_open:
                ruler_interval_len.append(
                    str(abs(big_ruler_intervals[i].right_value - big_ruler_intervals[i].left_value)))
                cnt += 1
            if cnt >= len(CR.left_dict):
                break
        tmp_pattern = "#".join(ruler_interval_len)
        current_the_most_left_ruler_interval = big_ruler_intervals[0]
        next_ruler_interval_len = CR.left_dict[tmp_pattern]
        new_ruler_interval = (Interval(current_the_most_left_ruler_interval.left_value - next_ruler_interval_len,
                                       current_the_most_left_ruler_interval.left_value, True, True))
        if new_ruler_interval.left_value > left_interval_range.left_value:
            big_ruler_intervals = [new_ruler_interval] + big_ruler_intervals
            continue
        elif new_ruler_interval.left_value == left_interval_range.left_value and new_ruler_interval.left_open == left_interval_range.left_open:
            big_ruler_intervals = [new_ruler_interval] + big_ruler_intervals
            continue
        else:
            break
    return big_ruler_intervals, starting_ruler_interval


def find_left_period(D, left_interval_range, CR):
    big_ruler_intervals, starting_ruler_interval = build_left_ruler_intervals(left_interval_range, CR)
    big_ruler_intervals = big_ruler_intervals[0: big_ruler_intervals.index(starting_ruler_interval) + 1]

    len_big_ruler_intervals = len(big_ruler_intervals)
    for i in range(len_big_ruler_intervals - 1, -1, -1):
        if big_ruler_intervals[i].left_open:
            continue
        first = big_ruler_intervals[i]
        try:
            second_index = big_ruler_intervals.index(Interval(first.left_value - CR.w, first.left_value - CR.w, False, False))
        except:
            break

        first_interval = big_ruler_intervals[second_index: i+1]
        for k in range(i - 1, -1, -1):
            if big_ruler_intervals[k].left_open:
                continue
            first = big_ruler_intervals[k]
            try:
                second_index = big_ruler_intervals.index(
                    Interval(first.left_value - CR.w, first.left_value - CR.w, False, False))
            except:
                break
            second_interval = big_ruler_intervals[second_index: k+1]
            if has_same_pattern(second_interval, first_interval) and has_same_facts(first_interval, second_interval,
                                                                                    D):
                varrho_left_dict = defaultdict(list)
                start_index = second_interval[0].left_value
                end_index = first_interval[0].left_value

                left_period = Interval(start_index, end_index, False, True)

                for predicate in D:
                    for entity in D[predicate]:
                        for ruler in D[predicate][entity]:
                            intersection_ruler = Interval.intersection(ruler, left_period)
                            if intersection_ruler:
                                varrho_left_dict[intersection_ruler].append(str(Atom(predicate, entity)))
                return left_period, varrho_left_dict

    return None, None


def build_right_ruler_intervals(right_interval_range, CR):
    big_ruler_intervals = CR.right_initial_ruler_intervals[:]
    base_index = big_ruler_intervals.index(Interval(CR.max_x, CR.max_x, False, False))
    starting_ruler_interval = big_ruler_intervals[base_index+2]
    while big_ruler_intervals[-1].right_value <= right_interval_range.right_value:
        ruler_interval_len = []
        cnt = 0
        if big_ruler_intervals[-1].left_open:
            new_ruler_interval = Interval(big_ruler_intervals[-1].right_value, big_ruler_intervals[-1].right_value,
                                          False, False)
            if new_ruler_interval.right_value < right_interval_range.right_value:
                big_ruler_intervals = big_ruler_intervals + [new_ruler_interval]
                continue
            elif new_ruler_interval.right_value == right_interval_range.right_value and new_ruler_interval.right_open == right_interval_range.right_open:
                big_ruler_intervals = big_ruler_intervals + [new_ruler_interval]
                continue
            else:
                break

        for i in range(len(big_ruler_intervals) - 1, -1, -1):
            if big_ruler_intervals[i].left_open:
                ruler_interval_len.append(
                    str(abs(big_ruler_intervals[i].right_value - big_ruler_intervals[i].left_value)))
                cnt += 1
            if cnt >= len(CR.right_dict):
                break

        tmp_pattern = "#".join(ruler_interval_len)
        current_the_most_right_ruler_interval = big_ruler_intervals[-1]
        next_ruler_interval_len = CR.right_dict[tmp_pattern]
        new_ruler_interval = (Interval(current_the_most_right_ruler_interval.right_value,
                                       current_the_most_right_ruler_interval.right_value + next_ruler_interval_len,
                                       True, True))
        if new_ruler_interval.right_value < right_interval_range.right_value:
            big_ruler_intervals = big_ruler_intervals + [new_ruler_interval]
        elif new_ruler_interval.right_value == right_interval_range.right_value and \
                new_ruler_interval.right_open == right_interval_range.right_open:
            big_ruler_intervals = big_ruler_intervals + [new_ruler_interval]
        elif new_ruler_interval.right_value == right_interval_range.right_value and not right_interval_range.right_open:
            big_ruler_intervals = big_ruler_intervals + [new_ruler_interval]
        else:
            break
    return big_ruler_intervals, starting_ruler_interval


def find_right_period(D, right_interval_range, CR):
    big_ruler_intervals,  starting_ruler_interval = build_right_ruler_intervals(right_interval_range, CR)
    big_ruler_intervals = big_ruler_intervals[big_ruler_intervals.index(starting_ruler_interval):]
    len_big_ruler_intervals = len(big_ruler_intervals)

    for i in range(0, len_big_ruler_intervals-1):
        if big_ruler_intervals[i].left_open:
            continue
        first = big_ruler_intervals[i]
        try:
            second_index = big_ruler_intervals.index(Interval(first.left_value + CR.w, first.left_value + CR.w, False, False))
        except:
            break
        first_interval = big_ruler_intervals[i:second_index+1]

        for k in range(i+1, len_big_ruler_intervals):
            if big_ruler_intervals[k].left_open:
                continue
            first = big_ruler_intervals[k]
            try:
                second_index = big_ruler_intervals.index(
                    Interval(first.left_value + CR.w, first.left_value + CR.w, False, False))
            except:
                break
            second_interval = big_ruler_intervals[k: second_index+1]
            if has_same_pattern(second_interval, first_interval) and has_same_facts(first_interval, second_interval, D):
                varrho_right_dict = defaultdict(list)
                start_index = first_interval[-1].right_value
                end_index = second_interval[-1].right_value
                right_period = Interval(start_index, end_index, True, False)

                for predicate in D:
                    for entity in D[predicate]:
                        for ruler in D[predicate][entity]:
                             intersection_ruler = Interval.intersection(ruler, right_period)
                             if intersection_ruler:
                                   varrho_right_dict[intersection_ruler].append(str(Atom(predicate, entity)))
                return right_period,  varrho_right_dict

    return None, None


def entail(fact, D):
    if fact.predicate not in D:
        return False
    else:
        if not fact.entity in D[fact.predicate]:
            return False
        else:
            intervals = D[fact.predicate][fact.entity]
            for interval in intervals:
                if Interval.inclusion(fact.interval, interval):
                    return True
            else:
                return False


def find_periods(CR, fact=None):
    left_period, left_len = defaultdict(list), 0
    right_period, right_len = defaultdict(list), 0
    cnt = 0
    number_mat = 0
    while True:
        common_fragment = CommonFragment(CR.base_interval)
        common_fragment.common = Interval(Decimal("-inf"), Decimal("+inf"), True, True)

        delta_new = naive_immediate_consequence_operator(D=CR.D, rules=CR.Program, D_index=CR.D_index)
        print("The number of mat:", number_mat)
        number_mat += 1
        diff_delta = []
        terminate_flag = False
        for head_predicate in delta_new:
            for head_entity, T in delta_new[head_predicate].items():
                if head_predicate not in CR.D or head_entity not in CR.D[head_predicate]:
                    diff_delta = T
                else:
                    for interval1 in T:
                        diff_delta += Interval.diff(interval1, CR.D[head_predicate][head_entity])

                for cr_interval in diff_delta:
                    if Interval.intersection(cr_interval, common_fragment.base_interval):
                        common_fragment.cr_flag = False
                        common_fragment.common = None
                        terminate_flag = True
                        break
                    else:
                        if cr_interval.right_value <= common_fragment.base_interval.left_value:
                            if cr_interval.right_value >= common_fragment.common.left_value:
                                common_fragment.common.left_value = cr_interval.right_value
                                common_fragment.common.left_open = not cr_interval.right_open
                        elif cr_interval.left_value >= common_fragment.base_interval.right_value:
                            if cr_interval.left_value <= common_fragment.common.right_value:
                                common_fragment.common.right_value = cr_interval.left_value
                                common_fragment.common.right_open = not cr_interval.left_open
                        else:
                            print(str(cr_interval))
                            print(str(common_fragment.common))
                            raise ValueError("Error Happen")
                if terminate_flag:
                    break

            if terminate_flag:
                break

        if len(diff_delta) == 0:
            # fixpoint
            common_fragment.common.left_value = decimal.Decimal("-inf")
            common_fragment.common.left_open = True
            common_fragment.common.right_value = decimal.Decimal("+inf")
            common_fragment.common.right_open = True
            return CR.D, common_fragment.common, None, None, None, None, None, None

        cnt += 1
        if fact is not None:
            if entail(fact, CR.D):
                print("The fact: {} is entailed".format(str(fact)))
                exit()

        if common_fragment.common is None:
            # add the new facts to the dataset
            for tmp_predicate in delta_new:
                for tmp_entity in delta_new[tmp_predicate]:
                    if tmp_predicate not in CR.D or tmp_entity not in CR.D[tmp_predicate]:
                        CR.D[tmp_predicate][tmp_entity] = CR.D[tmp_predicate][tmp_entity] + delta_new[tmp_predicate][tmp_entity]
                        # update index
                        for i, item in enumerate(tmp_entity):
                            CR.D_index[tmp_predicate][str(i) + "@" + item.name].append(tmp_entity)
                        if len(tmp_entity) > 2:
                            for i, item1 in enumerate(tmp_entity):
                                for j, item2 in enumerate(tmp_entity):
                                    if j <= i:
                                        continue
                                    CR.D_index[tmp_predicate][
                                        str(i) + "@" + item1.name + "||" + str(j) + "@" + item2.name].append(tmp_entity)
                    elif tmp_predicate in CR.D and tmp_entity in CR.D[tmp_predicate]:
                        CR.D[tmp_predicate][tmp_entity] += delta_new[tmp_predicate][tmp_entity]
            coalescing_d(CR.D)
            continue

        varrho_left_range = Interval(common_fragment.common.left_value, CR.min_x, common_fragment.common.left_open, True)
        varrho_right_range = Interval(CR.max_x, common_fragment.common.right_value, True, common_fragment.common.right_open)
        if varrho_left_range.left_value in [Decimal("-inf")] and varrho_right_range.right_value in [Decimal("+inf")]:
            return CR.D, common_fragment.common, None, None, None, None, None, None

        if varrho_left_range.left_value in [Decimal("-inf")]:
            varrho_right, varrho_right_dict = find_right_period(CR.D, varrho_right_range, CR)
            if varrho_right is not None:
                right_len = varrho_right.right_value - varrho_right.left_value
                for key, values in varrho_right_dict.items():
                    for value in values:
                        right_period[value].append(key)
                for key, value in right_period.items():
                    right_period[key] = coalescing(value)

                return CR.D, common_fragment.common, None, None, None, varrho_right, right_period, right_len
        else:
            varrho_left, varrho_left_dict = find_left_period(CR.D, varrho_left_range, CR)
            if varrho_left is not None:
                if varrho_right_range.right_value in [Decimal("+inf")]:
                    left_len = varrho_left.right_value - varrho_left.left_value
                    for key, values in varrho_left_dict.items():
                        for value in values:
                            left_period[value].append(key)
                    for key, value in left_period.items():
                        left_period[key] = coalescing(value)
                    return CR.D, common_fragment.common, varrho_left, left_period, left_len, None, None, None

                else:
                    varrho_right, varrho_right_dict = find_right_period(CR.D, varrho_right_range, CR)
                    if varrho_right is not None:
                        left_len = varrho_left.right_value - varrho_left.left_value
                        for key, values in varrho_left_dict.items():
                            for value in values:
                                left_period[value].append(key)
                        for key, value in left_period.items():
                            left_period[key] = coalescing(value)

                        right_len = varrho_right.right_value - varrho_right.left_value
                        for key, values in varrho_right_dict.items():
                            for value in values:
                                right_period[value].append(key)
                        for key, value in right_period.items():
                            right_period[key] = coalescing(value)
                        return CR.D, common_fragment.common, varrho_left, left_period, left_len, varrho_right, right_period, right_len

        for tmp_predicate in delta_new:
            for tmp_entity in delta_new[tmp_predicate]:
                if tmp_predicate not in CR.D or tmp_entity not in CR.D[tmp_predicate]:
                    CR.D[tmp_predicate][tmp_entity] = delta_new[tmp_predicate][tmp_entity]
                    # update index
                    for i, item in enumerate(tmp_entity):
                        CR.D_index[tmp_predicate][str(i) + "@" + item.name].append(tmp_entity)
                    if len(tmp_entity) > 2:
                        for i, item1 in enumerate(tmp_entity):
                            for j, item2 in enumerate(tmp_entity):
                                if j <= i:
                                    continue
                                CR.D_index[tmp_predicate][
                                    str(i) + "@" + item1.name + "||" + str(j) + "@" + item2.name].append(tmp_entity)
                elif tmp_predicate in CR.D and tmp_entity in CR.D[tmp_predicate]:
                    CR.D[tmp_predicate][tmp_entity] += delta_new[tmp_predicate][tmp_entity]

        coalescing_d(CR.D)


def fact_entailment(D, fact, base_interval, left_period, left_len, right_period, right_len):
    if fact.predicate not in D:
        return False
    else:
        if not fact.entity in D[fact.predicate]:
            return False
        else:
            intervals = D[fact.predicate][fact.entity]
            for interval in intervals:
                if Interval.inclusion(fact.interval, interval):
                    return True
            else:
                if Interval.inclusion(fact.interval, base_interval):
                    return False
                # using the canonical representation to do the checking
                elif fact.interval.left_value < base_interval.left_value and not left_period:
                    # less than the base interval and the left_period is empty
                    return False
                elif fact.interval.right_value > base_interval.right_value and not right_period:
                    # greater than the base interval and the right_period is empty
                    return False
                else:
                    target_interval = fact.interval
                    if target_interval.left_value >= base_interval.left_value:
                        if not right_period:
                            return False
                        last_interval = D[fact.predicate][fact.entity][-1]
                        remain_interval = Interval.diff(target_interval, [last_interval])
                        if len(remain_interval) != 1:
                            return False

                        else:
                            remain_interval = remain_interval[0]
                            repeated_intervals = right_period[str(Atom(fact.predicate, fact.entity))]
                            for interval in repeated_intervals:
                                if interval.right_value - interval.left_value == right_len:
                                    #infinity range
                                    return True
                                else:
                                    new_interval = Interval(interval.left_value + right_len * math.ceil((target_interval.right_value-interval.right_value)/right_len),
                                                            interval.right_value + right_len * math.ceil((target_interval.right_value-interval.right_value)/right_len),
                                                            interval.left_open, interval.right_open)
                                    if Interval.inclusion(remain_interval, new_interval):
                                        return True
                            return False

                    elif target_interval.right_value <= base_interval.right_value:
                        if not left_period:
                            return False
                        last_interval = D[fact.predicate][fact.entity][0]
                        remain_interval = Interval.diff(target_interval, [last_interval])
                        if len(remain_interval) != 1:
                            return False

                        else:
                            remain_interval = remain_interval[0]
                            repeated_intervals = left_period[str(Atom(fact.predicate, fact.entity))]
                            for interval in repeated_intervals:
                                if interval.right_value - interval.left_value == right_len:
                                    # infinity range
                                    return True
                                else:
                                    new_interval = Interval(
                                        interval.left_value - left_len * math.ceil(abs(target_interval.right_value-interval.left_value) / left_len),
                                        interval.right_value - left_len * math.ceil(abs(target_interval.right_value-interval.left_value) / left_len),
                                        interval.left_open, interval.right_open)
                                    if Interval.inclusion(remain_interval, new_interval):
                                        return True
                            return False
                    else:
                        if not left_period or not right_period:
                            return False

                        remain_interval = Interval.diff(target_interval, [D[fact.predicate][fact.entity][0]])
                        if len(remain_interval) != 1:
                            return False
                        remain_interval = remain_interval[0]
                        repeated_intervals = left_period[str(Atom(fact.predicate, fact.entity))]
                        for interval in repeated_intervals:
                            if interval.right_value - interval.left_value == right_len:
                                # infinity range
                               break
                            else:
                                new_interval = Interval(
                                    interval.left_value - left_len * math.ceil(abs(target_interval.left_value-interval.left_value) / left_len),
                                    interval.right_value - left_len * math.ceil(abs(target_interval.left_value-interval.left_value) / left_len),
                                    interval.left_open, interval.right_open)
                                if Interval.inclusion(remain_interval, new_interval):
                                    break
                        else:
                            return False

                        remain_interval = Interval.diff(target_interval, [D[fact.predicate][fact.entity][-1]])
                        if len(remain_interval) != 1:
                            return False

                        remain_interval = remain_interval[0]
                        repeated_intervals = right_period[str(fact.atom)]
                        for interval in repeated_intervals:
                            if interval.right_value - interval.left_value == right_len:
                                # infinity range
                                return True
                            else:
                                new_interval = Interval(
                                    interval.left_value + right_len * math.ceil((target_interval.right_value - interval.right_value) / right_len),
                                    interval.right_value + right_len * math.ceil((target_interval.right_value - interval.right_value) / right_len),
                                    interval.left_open, interval.right_open)
                                if Interval.inclusion(remain_interval, new_interval):
                                    return True
                        else:
                            return False








from meteor_reasoner.utils.ruler_interval import *
from meteor_reasoner.materialization.coalesce import *
from meteor_reasoner.materialization.index_build import *
from meteor_reasoner.classes.literal import Literal, BinaryLiteral
import decimal,copy
from meteor_reasoner.canonical.calculate_w import get_w

class CanonicalRepresentation:
    def __init__(self, D, Program):
        self.D = D
        self.D_index = defaultdict(lambda : defaultdict(list))
        self.Program = Program

    def is_bounded_program(self):
        for rule in self.Program:
            for literal in rule.body:
                if isinstance(literal, Literal):
                    for op in literal.operators:
                        if op.right_value in [decimal.Decimal("+inf")]:
                            return False
                elif isinstance(literal, BinaryLiteral):
                    if literal.operator.right_value in [decimal.Decimal("+inf")]:
                        return False
        return True

    def initilization(self):
        t_program = copy.deepcopy(self.Program)
        self.w = 2 * get_w(t_program)
        coalescing_d(self.D)
        build_index(self.D,  self.D_index)
        self.points, self.min_x, self.max_x = get_dataset_points_x(self.D, min_x_flag=True)
        self.base_interval = Interval(self.min_x, self.max_x, False, False)
        self.z, self.gcd = get_gcd(self.Program)
        self.left_initial_ruler_intervals, self.right_initial_ruler_intervals = get_initial_ruler_intervals(self.points[:], left_border= self.min_x-self.gcd,
                                                                      right_border=self.max_x+self.gcd, gcd=self.gcd)
        # self.left_initial_ruler_intervals = [
        #     Interval(self.min_x-self.gcd, self.min_x-self.gcd, False, False),
        #     Interval(self.min_x - self.gcd, self.min_x, True, True)] + self.left_initial_ruler_intervals
        # self.right_initial_ruler_intervals = self.right_initial_ruler_intervals + [Interval(self.max_x + self.gcd, self.max_x + self.gcd, False, False),
        #     Interval(self.max_x + self.gcd, self.max_x, True, True)
        # ]
        # self.initial_ruler_intervals =[
        #     Interval(self.min_x-self.gcd, self.min_x-self.gcd, False, False),
        #     Interval(self.min_x - self.gcd, self.min_x, True, True),
        #     Interval(self.min_x, self.min_x, False, False),
        #     Interval(self.max_x, self.max_x, False, False),
        #     Interval(self.max_x + self.gcd, self.max_x + self.gcd, False, False),
        #     Interval(self.max_x + self.gcd, self.max_x, True, True)
        # ]
        # start_time = time.time()
        self.left_dict, self.right_dict = construct_left_right_pattern(self.points, self.gcd)
        # self.left_dict = {"1.0": Decimal("1.0")}
        # self.right_dict = {"1.0": Decimal("1.0")}


from itertools import chain, combinations
from meteor_reasoner.automata.window import *
from meteor_reasoner.automata.satisfy import *
from meteor_reasoner.materialization.coalesce import *
from meteor_reasoner.automata.guess import *
from meteor_reasoner.materialization.utils import *
import decimal


class BuchiAutomata:
    def __init__(self, D, program, unbounded_literals, p_literals, constants, involved_predicates, automata_predicates,left_dict,
                 right_dict, x, z, gcd, points):
        self.D = D
        self.program = program
        self.x = x
        self.z = z
        self.gcd = gcd
        self.points = points
        self.p_literals = p_literals
        self.constants = constants

        self.unbounded_literals = unbounded_literals
        self.window_ruler_interval_list = []

        self.involved_predicates = involved_predicates
        self.automata_predicates = automata_predicates

        self.left_window_pattern_dict = left_dict
        self.right_window_pattern_dict = right_dict

        self.window_must_include_static = defaultdict(list)
        self.variable_literals = set()

        self.left_conditions = self.get_conditions("left")
        self.right_conditions = self.get_conditions("right")

    def get_conditions(self, d):
        F = set()
        for literal in self.unbounded_literals:
            if d == "left":
                if isinstance(literal, BinaryLiteral):
                    if literal.operator.interval.right_value == decimal.Decimal("inf") and literal.operator.name == "Since":
                        F.add(literal)
                else:
                    if isinstance(literal, Literal) and len(literal.operators) == 1 and (
                            literal.operators[0].interval.right_value == decimal.Decimal("inf")) and literal.operators[0].name == "Boxminus":
                        F.add(literal)

            else:
                if isinstance(literal, BinaryLiteral):
                    if literal.operator.interval.right_value == decimal.Decimal("inf") and literal.operator.name == "Until":
                        F.add(literal)
                else:
                    if isinstance(literal, Literal) and len(literal.operators) == 1 and (
                            literal.operators[0].interval.right_value == decimal.Decimal("inf")) and literal.operators[0].name == "Boxplus":
                        F.add(literal)
        return F



    def build_prior(self, must_literals):
        initial_window_len, initial_window_ruler_intervals = get_initial_ruler_intervals(self.points,
                                                                                         left_border=-2 * self.x - 2 * self.z,
                                                                                         right_border=2 * self.x + 2 * self.z,
                                                                                         gcd=self.gcd)
        self.window_ruler_interval_list = initial_window_ruler_intervals
        for ruler_interval in initial_window_ruler_intervals:
            self.window_must_include_static[ruler_interval] = []

        for ruler_interval in initial_window_ruler_intervals:
            for literal in must_literals:
                if interval_intesection_intervallist(ruler_interval, must_literals[literal]):
                    self.window_must_include_static[ruler_interval].append(literal)
                if isinstance(literal, BinaryLiteral):
                    if literal.left_literal.get_predicate() in self.automata_predicates or literal.right_literal.get_predicate() in self.automata_predicates:
                        self.variable_literals.add(literal)
                else:
                    if literal.get_predicate() in self.automata_predicates:
                        self.variable_literals.add(literal)

            for predicate in self.involved_predicates:
                for entity in self.D[predicate]:
                    if interval_intesection_intervallist(ruler_interval, self.D[predicate][entity]):
                        self.window_must_include_static[ruler_interval].append(
                            Atom(predicate, entity))
                    if predicate in self.automata_predicates:
                        self.variable_literals.add(Atom(predicate, entity))
        program_variables = set()
        for rule in self.program:
            head = rule.head
            if literal_contain_no_variable(head):
                program_variables.add(head)

            for literal in rule.body:
               if literal_contain_no_variable(literal):
                   program_variables.add(literal)

               elif isinstance(literal, Atom):
                   if literal.get_predicate() not in ["Bottom", "Top"]:
                       for com_constants in combinations(self.constants, len(literal.entity)):
                           tmp_dict = {}
                           tmp_literal = copy.deepcopy(literal)
                           for i in range(len(tmp_literal.entity)):
                               if literal.entity[i].type == "variable":
                                   if literal.entity[i].name not in tmp_dict:
                                       tmp_literal.entity[i].name = com_constants[i]
                                       tmp_literal.entity[i].type = "constant"
                                       tmp_dict[literal.entity[i].name] = com_constants[i]
                                   else:
                                       tmp_literal.entity[i].name = tmp_dict[literal.entity[i].name]
                           program_variables.add(tmp_literal)


               elif isinstance(literal, Literal):
                   if literal.get_predicate() not in ["Bottom", "Top"]:
                       for com_constants in combinations(self.constants, len(literal.atom.entity)):
                           tmp_dict = {}
                           tmp_atom = copy.deepcopy(literal.atom)
                           for i in range(len(tmp_atom.entity)):
                               if tmp_atom.entity[i].type == "variable":
                                   if tmp_atom.entity[i].name not in tmp_dict:
                                       tmp_atom.entity[i].name = com_constants[i]
                                       tmp_atom.entity[i].type = "constant"
                                       tmp_dict[tmp_atom.entity[i].name] = com_constants[i]
                                   else:
                                       tmp_atom.entity[i].name = tmp_dict[tmp_atom.entity[i].name]
                           tmp_literal = copy.deepcopy(literal)
                           tmp_literal.atom  = tmp_atom
                           program_variables.add(tmp_literal)

               elif isinstance(literal, BinaryLiteral):
                    left_literal = literal.left_literal
                    right_literal = literal.right_literal
                    if left_literal.get_predicate() not in ["Bottom", "Top"]:
                        for left_com_constants in combinations(self.constants, len(left_literal.entity)):
                            tmp_dict = {}
                            left_tmp_atom = copy.deepcopy(left_literal)
                            for i in range(len(left_tmp_atom.entity)):
                                if left_tmp_atom.entity[i].type == "variable":
                                    if left_tmp_atom.entity[i].name not in tmp_dict:
                                        left_tmp_atom.entity[i].name = left_com_constants[i]
                                        left_tmp_atom.entity[i].type = "constant"
                                        tmp_dict[left_tmp_atom.entity[i].name] = left_com_constants[i]
                                    else:
                                        left_tmp_atom.entity[i].name = tmp_dict[left_tmp_atom.entity[i].name]
                            for right_com_constants in combinations(self.constants, len(left_literal.entity)):
                                right_tmp_atom = copy.deepcopy(right_literal)
                                for i in range(len(right_tmp_atom.entity)):
                                    if right_tmp_atom.entity[i].type == "variable":
                                        if right_tmp_atom.entity[i].name not in tmp_dict:
                                            right_tmp_atom.entity[i].name = right_com_constants[i]
                                            right_tmp_atom.entity[i].type = "constant"
                                            tmp_dict[right_tmp_atom.entity[i].name] = right_com_constants[i]
                                        else:
                                            right_tmp_atom.entity[i].name = tmp_dict[right_tmp_atom.entity[i].name]

                                tmp_literal = copy.deepcopy(literal)
                                tmp_literal.left_literal = left_tmp_atom
                                tmp_literal.right_literal = right_tmp_atom
                                program_variables.add(tmp_literal)

        self.variable_literals = self.variable_literals | set(self.unbounded_literals) | program_variables
        atm_btm = Atom("Bottom", tuple([Term("nan")]))
        atm_top = Atom("Top", tuple([Term("nan")]))
        literal_btm = Literal(atm_btm, [])
        literal_top = Literal(atm_top, [])
        for atm in [atm_btm, atm_top, literal_btm, literal_top]:
            if atm in self.variable_literals:
                self.variable_literals.remove(atm)
       
        
    def get_unbounded_literals_for_window(self, atom, ruler_interval, direction):
        """
        Return a list of possible unbounded literals for the given ruler_interval based
        on the pre-materialized results in [-2x-2z, 2x+2z]
        Args:
            atom (Atom instance):
            ruler_interval (Interval instance):
            direction (str): ``left'' or ``right''

        Returns:
            list of Literal instances

        """
        if ruler_interval not in self.window_must_include_static:
            return []
        unbounded_literals = []
        start_i = self.window_ruler_interval_list.index(ruler_interval)
        if direction == "left":
            for i in range(start_i, -1, -1):
                next_ruler_interval = self.window_ruler_interval_list[i]
                if atom not in self.window_must_include_static[next_ruler_interval]:
                    break
            else:
                open_operator = Operator("Boxminus", Interval(decimal.Decimal(0), decimal.Decimal("inf"), True, True))
                closed_operator = Operator("Boxminus", Interval(decimal.Decimal(0), decimal.Decimal("inf"), False, True))
                open_unbounded_literal = Literal(atom, [open_operator])
                close_unbounded_literal = Literal(atom, [closed_operator])
                unbounded_literals.append(open_unbounded_literal)
                unbounded_literals.append(close_unbounded_literal)
                for rule in self.program:
                    if len(rule.body) == 1 and rule.body[0] in [close_unbounded_literal, open_unbounded_literal]:
                        unbounded_literals.append(rule.head)

        else:
            for i in range(start_i, len(self.window_ruler_interval_list)):
                next_ruler_interval = self.window_ruler_interval_list[i]
                if atom not in self.window_must_include_static[next_ruler_interval]:
                    break
            else:
                open_operator = Operator("Boxplus", Interval(decimal.Decimal(0), decimal.Decimal("inf"), True, True))
                closed_operator = Operator("Boxplus", Interval(decimal.Decimal(0), decimal.Decimal("inf"), False, True))
                open_unbounded_literal = Literal(atom, [open_operator])
                close_unbounded_literal = Literal(atom, [closed_operator])
                unbounded_literals.append(open_unbounded_literal)
                unbounded_literals.append(close_unbounded_literal)
                for rule in self.program:
                    if len(rule.body) == 1 and rule.body[0] in [close_unbounded_literal, open_unbounded_literal]:
                        unbounded_literals.append(rule.head)

        return unbounded_literals

    def get_literals_for_ruler_interval(self, ruler_interval, direction, last_ruler_interval_literals=None):
        """
        Return a list of Atom or Literal instances for a given ruler interval
        Args:
            ruler_interval (an Interval instance):

        Returns:
            list of Atoms or Literal instances

        """
        static_must_literals = set(self.window_must_include_static[ruler_interval])
        unbounded_must_literals = set()
        if last_ruler_interval_literals is not None:
            if direction == "left":
                for literal in last_ruler_interval_literals:
                    if isinstance(literal, Literal) and len(literal.operators) == 1 and literal.operators[ 0].name == "Boxminus" and\
                            literal.operators[0].interval.right_value == decimal.Decimal("inf"):
                        unbounded_must_literals.add(literal)
                        unbounded_must_literals.add(literal.atom)
            else:
                for literal in last_ruler_interval_literals:
                    if isinstance(literal, Literal) and len(literal.operators) == 1 and literal.operators[ 0].name == "Boxplus" and \
                            literal.operators[0].interval.right_value == decimal.Decimal("inf"):

                        unbounded_must_literals.add(literal)
                        unbounded_must_literals.add(literal.atom)


        for com_literals in chain.from_iterable(combinations(self.variable_literals, r)
                                                   for r in range(len(self.variable_literals) + 1)):

            yield set(com_literals) | static_must_literals | unbounded_must_literals

    def get_literals_for_ruler_interval_with_visited(self, ruler_interval, direction, visited, last_ruler_interval_literals=None):
        """
        Return a list of Atom or Literal instances for a given ruler interval
        Args:
            ruler_interval (an Interval instance):

        Returns:
            list of Atoms or Literal instances

        """
        static_must_literals = set(self.window_must_include_static[ruler_interval])
        unbounded_must_literals = set()
        for literal in static_must_literals:
            if isinstance(literal, Atom): # and literal not in visited:
                if direction == "left":
                    res = self.get_unbounded_literals_for_window(literal, ruler_interval, "left")
                    if len(res) > 0:
                        visited[literal] = 1
                        unbounded_must_literals |= set(res)
                else:
                    res = self.get_unbounded_literals_for_window(literal, ruler_interval, "right")
                    if len(res) > 0:
                        visited[literal] = 1
                        unbounded_must_literals |= set(res)

        if last_ruler_interval_literals is not None:
            if direction == "left":
                for literal in last_ruler_interval_literals:
                    if isinstance(literal, Literal) and len(literal.operators) == 1 and literal.operators[0].name == "Boxminus" and \
                            literal.operators[0].interval.right_value == decimal.Decimal("inf"):
                        unbounded_must_literals.add(literal)
                        unbounded_must_literals.add(literal.atom)
            else:
                for literal in last_ruler_interval_literals:
                    if isinstance(literal, Literal) and len(literal.operators) == 1 and literal.operators[0].name == "Boxplus" and \
                            literal.operators[0].interval.right_value == decimal.Decimal("inf"):
                        unbounded_must_literals.add(literal)
                        unbounded_must_literals.add(literal.atom)

        # prunning
        pruning_elements = set([atom for atom in self.variable_literals if not isinstance(atom, BinaryLiteral) and atom.get_predicate() == "New"])
        # print("Variable literals:")
        # print([str(literal) for literal in self.variable_literals])
        if ruler_interval in self.window_ruler_interval_list:
            for literal in self.variable_literals:
                if isinstance(literal, Atom):
                    if literal not in self.window_must_include_static[ruler_interval]:
                        pruning_elements.add(literal)
                        u_literal = Literal(literal, [Operator("Boxminus", Interval(0, Decimal("inf"), False, True))])
                        pruning_elements.add(u_literal)
                        u_literal = Literal(literal, [Operator("Boxplus", Interval(0, Decimal("inf"), False, True))])
                        pruning_elements.add(u_literal)
                        for p_literal in self.p_literals:
                            if isinstance(p_literal, Literal) and len(p_literal.operators) == 1 and \
                                    p_literal.operators[0].interval.left_value == p_literal.operators[0].interval.right_value == 0 and p_literal.atom == literal:
                                pruning_elements.add(p_literal)

        left_variable_predicate = self.variable_literals - unbounded_must_literals - static_must_literals - pruning_elements
        for com_literals in chain.from_iterable(combinations(left_variable_predicate, r)
                                                   for r in range(len(left_variable_predicate) + 1)):
            bounded_fix = set()
            res = set(com_literals) | static_must_literals | unbounded_must_literals
            for literal in res:
                if isinstance(literal, Literal) and len(literal.operators) == 1 and literal.operators[0].name in  ["Boxminus", "Boxplus"] and \
                    literal.operators[0].interval.right_value == decimal.Decimal("inf"):
                    if literal.operators[0].interval.left_value == 0 and literal.operators[0].interval.left_open and literal.atom in res:
                        new_bounded = copy.deepcopy(literal)
                        new_bounded.operators[0].left_open = False
                        if new_bounded not in res:
                            bounded_fix.add(new_bounded)
                    elif literal.operators[0].interval.left_value == 0 and not literal.operators[0].interval.left_open:
                        new_bounded = copy.deepcopy(literal)
                        new_bounded.operators[0].left_open = True
                        if new_bounded not in res:
                            bounded_fix.add(new_bounded)

            yield res | bounded_fix

    def get_literals_for_window(self, window_ruler_interval_list,  visited, direction):
        def helper(index, result):
            if index == len(window_ruler_interval_list):
                window_ruler_intervals_literals = dict()
                for key, value in result.items():
                    window_ruler_intervals_literals[key] = set(value)

                yield AutomataWindow(window_ruler_interval_list, window_ruler_intervals_literals,
                                     self.left_window_pattern_dict, self.right_window_pattern_dict)
                return

            else:
                last_interval_literals = None
                if index != 0:
                    last_interval_literals = result[window_ruler_interval_list[index-1]]

                for com_literals in self.get_literals_for_ruler_interval_with_visited(
                        window_ruler_interval_list[index], direction, visited, last_interval_literals):
                    yield from helper(index+1, {**result, **{window_ruler_interval_list[index]: list(com_literals)}})

        yield from helper(0, {})

    def check_inconsistency(self):
        """
        This function check whether the given D and Program are consistency by running Buichi Automata.
        Returns:
           boolean
        """
        initial_window_len, initial_window_ruler_intervals =get_initial_ruler_intervals(
            [- self.x - self.z, -self.x + self.z])
        for l_min, r_min in self.search_window(initial_window_ruler_intervals):
            if self.search(l_min, "left", self.F_left):
                if self.search(r_min, "right", self.F_right):
                    return True
        return False

    def search_window(self, initial_window_ruler_intervals):
        """
        Instead of directly checking the window=[-x-z, -x+z, we start from
        the minimum window=[-x-z, -x+z], and then move in a rightward direction
        at a gcd step until the window reach to x+z. Then return two windows:
        lmin=[-x-z, -x+z], rmin=[x-z, x+z]
        Args:
            initial_window_ruler_intervals (list of Interval instances):

        Returns:
            (AutomataWindow instance, AutomataWindow instance)
        """
        visited = dict()
        def guess_help(w, w0):
            while w.ruler_intervals[-1] != Interval(self.x+self.z, self.x+self.z, False, False):
                last_interval_literals = w.ruler_intervals_literals[w.ruler_intervals[-1]]
                w.move_right()
                last_interval = w.ruler_intervals[-1]
                for com_literals in self.get_literals_for_ruler_interval_with_visited(last_interval, "right", visited, last_interval_literals):
                    w.ruler_intervals_literals[last_interval] = com_literals
                    bottom_flag = False
                    for ruler_interval in w.ruler_intervals:
                        must_installed_heads = set()
                        for rule in self.program:
                            must_installed_heads |= return_must_heads(rule, installed_literals=w.ruler_intervals_literals[ruler_interval])
                        w.ruler_intervals_literals[ruler_interval] |= must_installed_heads
                        if Atom("Bottom") in must_installed_heads:
                            bottom_flag = True
                            break

                    # begin: adding some must-be literals
                    add_p_literals(self.p_literals, w)
                    #end: adding some must-be literals

                    if bottom_flag:
                        continue

                    if not self.check_satisfy(w):
                        return None
                    else:
                        break
            return (w0, w)

        # This is a recursive version
        def helper(stack):
            if len(stack) > 0 and stack[-1].ruler_intervals[-1] == Interval(self.x+self.z, self.x+self.z, False, False):
                yield stack[0], stack[-1]
                return
            last_w = stack[-1]
            new_w = copy.deepcopy(last_w)
            last_interval_literals = new_w.ruler_intervals_literals[new_w.ruler_intervals[-1]]
            new_w.move_right()
            last_interval = new_w.ruler_intervals[-1]

            for com_literals in self.get_literals_for_ruler_interval_with_visited(last_interval, "right", visited, last_interval_literals):
                new_w.ruler_intervals_literals[last_interval] |= com_literals
                bottom_flag = False
                for ruler_interval in new_w.ruler_intervals:
                    must_installed_heads = set()
                    for rule in self.program:
                        must_installed_heads |= return_must_heads(rule, installed_literals=new_w.ruler_intervals_literals[
                            ruler_interval])
                    new_w.ruler_intervals_literals[ruler_interval] |= must_installed_heads
                    if Atom("Bottom") in must_installed_heads:
                        bottom_flag = True
                        break
                if bottom_flag:
                    continue
                # begin: adding some must-be literals
                add_p_literals(self.p_literals, new_w)
                # end: adding some must-be literals

                if self.check_satisfy(new_w):
                    yield from helper(stack+[new_w])

        flag = True
        for w in self.get_literals_for_window(initial_window_ruler_intervals, visited, "right"):
            bottom_flag = False
            for ruler_interval in w.ruler_intervals:
                must_installed_heads = set()
                for rule in self.program:
                    must_installed_heads |= return_must_heads(rule, installed_literals= w.ruler_intervals_literals[
                        ruler_interval])
                    w.ruler_intervals_literals[ruler_interval] |= must_installed_heads
                if Atom("Bottom") in must_installed_heads:
                    bottom_flag = True
                    break
            if bottom_flag:
                continue

            # begin: adding some must-be literals
            add_p_literals(self.p_literals, w)
            # end: adding some must-be literals

            if self.check_satisfy(w):
                if flag:
                    res = guess_help(copy.deepcopy(w), copy.deepcopy(w))
                    flag = False
                    if res is not None:
                        yield res[0], res[1]
                    else:
                        yield from helper([w])
                else:
                    yield from helper([w])

    def consistency_check(self):
        """
        Returns:
           boolean
        """
        # 2z width
        initial_window_len, initial_window_ruler_intervals = get_initial_ruler_intervals(self.points,
            left_border=-self.x - self.z, right_border= -self.x + self.z, gcd=self.gcd)
        for l_min, r_min in self.search_window(initial_window_ruler_intervals):
            left_flag = self.search(l_min, "left", self.left_conditions)
            if left_flag:
                right_flag = self.search(r_min, "right", self.right_conditions)
                if right_flag:
                    return True
        return False

    def search(self, W_0, d, F):
        T = list()
        R = set()
        label = defaultdict(set)
        T.append(W_0)

        while len(T) > 0:
            W = T[-1]
            increase_flag = True
            while increase_flag:
                for tmp_interval in W.ruler_intervals:
                    if len(W.ruler_intervals_literals[tmp_interval]) != 0:
                        break
                else:
                    return True
                if d == "left":
                    W_prime = copy.deepcopy(W)
                    last_interval_literals = W_prime.ruler_intervals_literals[W_prime.ruler_intervals[-1]]
                    W_prime.move_left()
                    first_interval = W_prime.ruler_intervals[0]
                    for com_literals in self.get_literals_for_ruler_interval(first_interval,d, last_interval_literals):
                        W_prime.ruler_intervals_literals[first_interval] = com_literals

                        bottom_flag = False
                        for ruler_interval in W_prime.ruler_intervals:
                            must_installed_heads = set()
                            for rule in self.program:
                                must_installed_heads |= return_must_heads(rule, installed_literals=
                                W_prime.ruler_intervals_literals[
                                    ruler_interval])
                            W_prime.ruler_intervals_literals[ruler_interval] |= must_installed_heads
                            if Atom("Bottom") in must_installed_heads:
                                bottom_flag = True
                                break
                        if bottom_flag:
                            continue

                        # begin: adding some must-be literals
                        add_p_literals(self.p_literals, W_prime)
                        # end: adding some must-be literals

                        if self.check_satisfy(W_prime) and W_prime not in T + list(R):
                            T.append(W_prime)
                            label[W_prime] = set()
                            W = W_prime
                            break
                    else:
                        increase_flag = False

                else:
                    W_prime = copy.deepcopy(W)
                    last_interval_literals = W_prime.ruler_intervals_literals[W_prime.ruler_intervals[-1]]
                    W_prime.move_right()
                    last_interval = W_prime.ruler_intervals[-1]
                    for com_literals in self.get_literals_for_ruler_interval(last_interval, d, last_interval_literals):
                        W_prime.ruler_intervals_literals[last_interval] = com_literals
                        bottom_flag = False
                        for ruler_interval in W_prime.ruler_intervals:
                            must_installed_heads = set()
                            for rule in self.program:
                                must_installed_heads |= return_must_heads(rule, installed_literals=
                                W_prime.ruler_intervals_literals[
                                    ruler_interval])
                            W_prime.ruler_intervals_literals[ruler_interval] |= must_installed_heads
                            if Atom("Bottom") in must_installed_heads:
                                bottom_flag = True
                                break
                        if bottom_flag:
                            continue

                        # begin: adding some must-be literals
                        add_p_literals(self.p_literals, W_prime)
                        # end: adding some must-be literals
                        if self.check_satisfy(W_prime) and W_prime not in T + list(R):
                            T.append(W_prime)
                            label[W_prime] = set()
                            W = W_prime
                            break
                        else:
                            increase_flag = False
                            break
                    else:
                        increase_flag = False

            subset_conditions = set()
            if (W in label and len(label[W]) != 0) or window_contain_accepting_conditions(F, W, subset_conditions, d):
                label[W] = label[W].union(subset_conditions)
                if label[W] == F:
                    return True
                self.propagate([W], label[W], T, R, label, d)
                if label[W] == F:
                    return True
            R.add(W)
            T.pop()

        return False

    def propagate(self, N, P, T, R, label, d):
        while len(N) != 0:
            W = N.pop()
            if d == "left":
                W_prime = copy.deepcopy(W)
                last_interval_literals = W_prime.ruler_intervals_literals[W_prime.ruler_intervals[-1]]
                W_prime.move_left()

                first_interval = W_prime.ruler_intervals[0]
                for subset_literal in self.get_literals_for_ruler_interval(first_interval, d, last_interval_literals):
                    W_prime.ruler_intervals_literals[first_interval] = subset_literal

                    bottom_flag = False
                    for ruler_interval in W_prime.ruler_intervals:
                        must_installed_heads = set()
                        for rule in self.program:
                            must_installed_heads |= return_must_heads(rule, installed_literals=
                            W_prime.ruler_intervals_literals[
                                ruler_interval])
                        W_prime.ruler_intervals_literals[ruler_interval] |= must_installed_heads
                        if Atom("Bottom") in must_installed_heads:
                            bottom_flag = True
                            break
                    if bottom_flag:
                        continue

                    # begin: adding some must-be literals
                    add_p_literals(self.p_literals, W_prime)
                    # end: adding some must-be literals
                    if self.check_satisfy(W_prime) and W_prime in T + list(R):
                        if self.can_contribute(P, label[W_prime]):
                            N.append(W_prime)
                            label[W_prime] = label[W_prime].union(P)
            else:
                W_prime = copy.deepcopy(W)
                last_interval_literals = W_prime.ruler_intervals_literals[W_prime.ruler_intervals[-1]]
                W_prime.move_right()

                last_interval = W_prime.ruler_intervals[-1]
                for subset_literal in self.get_literals_for_ruler_interval(last_interval, d, last_interval_literals):
                    W_prime.ruler_intervals_literals[last_interval] = subset_literal

                    bottom_flag = False
                    for ruler_interval in W_prime.ruler_intervals:
                        must_installed_heads = set()
                        for rule in self.program:
                            must_installed_heads |= return_must_heads(rule, installed_literals=
                            W_prime.ruler_intervals_literals[ruler_interval])
                        W_prime.ruler_intervals_literals[ruler_interval] |= must_installed_heads
                        if Atom("Bottom") in must_installed_heads:
                            bottom_flag = True
                            break

                    if bottom_flag:
                        continue

                    # begin: adding some must-be literals
                    add_p_literals(self.p_literals, W_prime)
                    # end: adding some must-be literals
                    if self.check_satisfy(W_prime) and W_prime in T + list(R):
                        if self.can_contribute(P, label[W_prime]):
                            N.append(W_prime)
                            label[W_prime] = label[W_prime].union(P)

    def can_contribute(self, a, b):
        for item in a:
            if item not in b:
                return True
        return False

    def check_satisfy(self, w):
        if not check_satisfy_dataset(w, self.D):
            return False

        ruler_intervals_literals = w.ruler_intervals_literals
        ruler_intervals = w.ruler_intervals

        guess_D = defaultdict(lambda: defaultdict(list))
        binary_literals = []

        for ruler_interval in ruler_intervals:
            for literal in ruler_intervals_literals[ruler_interval]:
                if isinstance(literal, Atom) or (isinstance(literal, Literal) and len(literal.operators) == 0):
                    guess_D[literal.get_predicate()][literal.get_entity()].append(ruler_interval)
                elif isinstance(literal, BinaryLiteral):
                    binary_literals.append(literal)

                elif isinstance(literal, Literal):
                    if literal.operators[0].name == "Boxplus":
                        interval = Interval.add(ruler_interval, literal.operators[0].interval)
                        if literal.get_predicate() not in guess_D:
                            guess_D[literal.get_predicate()][literal.get_entity()] = [interval]
                        else:
                            guess_D[literal.get_predicate()][literal.get_entity()].append(interval)
                    elif literal.operators[0].name == "Boxminus":
                        interval = Interval.sub(ruler_interval, literal.operators[0].interval)
                        if literal.get_predicate() not in guess_D:
                            guess_D[literal.get_predicate()][literal.get_entity()] = [interval]
                        else:
                            guess_D[literal.get_predicate()][literal.get_entity()].append(interval)
                else:
                    raise Exception("Please convert Diamondminus or Diamondplus to Boxminus or Boxplus")

        if len(binary_literals) != 0:
            guess_points, left_right_literals = generate_guess_points(w, binary_literals, ruler_intervals, self.z, self.gcd)
            # guess existing intervals
            for literal in left_right_literals:
                if literal.get_predicate() in self.D and literal.get_entity() is None:
                    guess_D[literal.get_predicate()] = self.D[literal.get_predicate()]
                elif literal.get_predicate() in self.D and literal.get_entity() in self.D[literal.get_predicate()]:
                    guess_D[literal.get_predicate()][literal.get_entity()] = self.D[literal.get_predicate()][literal.get_entity()]
                else:
                    # guess an interval from guess_points
                    pass

        coalescing_d(guess_D)
        for literal in self.p_literals:
            if guess_D is None:
                continue
            T = apply(literal, guess_D)
            if len(T) == 0:
                for ruler_interval in ruler_intervals:
                    if literal in ruler_intervals_literals[ruler_interval]:
                        return False
            else:
                for ruler_interval in ruler_intervals:
                    if interval_intesection_intervallist(ruler_interval, T) and literal not in \
                            ruler_intervals_literals[ruler_interval]:
                        return False

        prgram_flag = check_satisfy_program(w, program=self.program)
        return prgram_flag

































from meteor_reasoner.graphutil.graph_strengthening import *
from meteor_reasoner.graphutil.multigraph import *


def finite_materialisability(rules):
    paris = construct_pair(rules)
    G = TemporalDependencyGraph()
    for pair in paris:
        G.add_edge(pair[0], pair[1], [pair[2], pair[3]])
    return not G.is_cyclic()






from meteor_reasoner.graphutil.graph_strengthening import *
from meteor_reasoner.graphutil.multigraph import *
from meteor_reasoner.classes import *
from meteor_reasoner.utils.loader import load_program


def test_finite_materialisability():

    rules = load_program(["A(X):-Boxplus[4]C(X)", "Boxplus[5]C(X):- A(X)"])
    rules = transformation(rules)
    print(rules[0])
    paris = construct_pair(rules)
    G = TemporalDependencyGraph()
    for pair in paris:
        G.add_edge(pair[0], pair[1], [pair[2], pair[3]])
    print(rule, ":")
    print(G.is_cyclic())


test_finite_materialisability()




from meteor_reasoner.utils.conversion import *
from meteor_reasoner.automata.buichi_automata import *
from meteor_reasoner.automata.utils import *
from meteor_reasoner.utils.normalize import *
from meteor_reasoner.materialization.materialize import *


def consistency(D_1, program, F=None):
    contain_bottom = False
    for rule in program:
        if rule.head.get_predicate() == "Bottom":
            contain_bottom = True
            break
    if F is None and not contain_bottom:
        return True

    if F is None:
        must_literals = defaultdict(list)
        coalescing_d(D_1)
        relevant_rules = program
        CF = CycleFinder(program=program)
        non_recursive_predicates = CF.get_non_recursive_predicates()
        involved_predicates = [predicate for predicate in CF.predicate2id if predicate not in ["Bottom", "Top"]]
        automata_predicates = list(set(involved_predicates) - set(non_recursive_predicates))

        d_atoms, d_unbounded_literals = extract_dataset(D_1, involved_predicates, automata_predicates)
        points, x = get_dataset_points_x(D_1)
        z, gcd = get_gcd(relevant_rules)

        left_dict, right_dict = construct_left_right_pattern(points, gcd)
        limit = Interval(-2 * z - 2 * x, 2 * x + 2 * z, False, False)
        while True:
            flag, delta_new = materialize(D_1, relevant_rules, K=1)
            has_intersection = False
            for predicate in delta_new:
                for entity in delta_new[predicate]:
                    for interval in delta_new[predicate][entity]:
                        if Interval.inclusion(interval, limit):
                            has_intersection = True
                            break
                    if has_intersection:
                        break
                if has_intersection:
                    break
            else:
                break

        p_literals = []
        for literal in must_literals.keys():
            if not isinstance(literal, Atom):
                p_literals.append(literal)

        automata = BuchiAutomata(D=D_1, program=relevant_rules,
                                 p_literals=p_literals,
                                 unbounded_literals=d_unbounded_literals,
                                 involved_predicates=involved_predicates,
                                 automata_predicates=automata_predicates,
                                 left_dict=left_dict,
                                 right_dict=right_dict,
                                 x=x, z=z, gcd=gcd, points=points)

        automata.build_prior(must_literals=must_literals)
        flag = automata.consistency_check()
        return flag

    else:
        must_literals = defaultdict(list)
        coalescing_d(D_1)

        # make the conversion
        new_rule, new_atom, _ = euqal_conversion(F)
        program.append(new_rule)
        D_1[new_atom.predicate][new_atom.entity] = [new_atom.interval]
        program = normalize(program)

        CF = CycleFinder(program=program)
        relevant_rules = CF.get_revevant_rules(F.get_predicate())
        CF = CycleFinder(program=relevant_rules)
        non_recursive_predicates = CF.get_non_recursive_predicates()
        involved_predicates = [predicate for predicate in CF.predicate2id if predicate not in ["Bottom", "Top"]]
        automata_predicates = list(set(involved_predicates) - set(non_recursive_predicates))


        D = defaultdict(lambda: defaultdict(dict))
        for predicate in involved_predicates:
            if predicate in D_1:
                for entity in D_1[predicate]:
                     D[predicate][entity] = D_1[predicate][entity]

        d_atoms, d_unbounded_literals = extract_dataset(D, involved_predicates, automata_predicates)
        points, x = get_dataset_points_x(D)
        z, gcd = get_gcd(relevant_rules)

        left_dict, right_dict = construct_left_right_pattern(points, gcd)
        limit = Interval(-2 * z - 2 * x, 2 * x + 2 * z, False, False)
        while True:
            flag, delta_new = materialize(D, relevant_rules, K=1)
            for predicate in delta_new:
                for entity in delta_new[predicate]:
                    for interval in delta_new[predicate][entity]:
                        if Interval.inclusion(interval, limit):
                            break
            else:
                break

        p_literals = []
        for literal in must_literals.keys():
            if not isinstance(literal, Atom):
                p_literals.append(literal)

        automata = BuchiAutomata(D=D, program=relevant_rules,
                                 p_literals=p_literals,
                                 unbounded_literals=d_unbounded_literals,
                                 involved_predicates=involved_predicates,
                                 automata_predicates=automata_predicates,
                                 left_dict=left_dict,
                                 right_dict=right_dict,
                                 x=x, z=z, gcd=gcd, points=points)

        automata.build_prior(must_literals=must_literals)
        flag = automata.consistency_check()
        return flag


from meteor_reasoner.classes.literal import *
from meteor_reasoner.classes.rule import *
from meteor_reasoner.classes.atom import Atom
from meteor_reasoner.classes.interval import Interval
from meteor_reasoner.classes.term import Term

def euqal_conversion(f):
    """
    This function is to convert the fact entailment program to the consistency checking problem by adding
    a new fact and a new rule.
    Args:
        PI (list of Rule instances):
        D (dictionary of dictionary): store all facts
        f (a query):

    Returns:
        newly-createdly Literal instance

    """
    t1, t2 = f.interval.left_value, f.interval.right_value
    head = Atom("Bottom", tuple([Term("nan")]))
    literal1 = Atom("New", f.entity)
    literal2 = Literal(Atom(f.predicate, f.entity))
    operator1 = Operator("Boxplus", Interval(0, t2-t1, False, f.interval.right_open))
    literal2.operators.append(operator1)
    rule = Rule(head, [literal1, literal2])
    fact = Atom("New", f.entity, Interval(t1, t1, False, False))
    return rule, fact, literal2



from meteor_reasoner.utils.parser import *
from collections import defaultdict
import time, os, datetime
import csv

def load_dataset(file_or_path):
    """
    Read string-like facts into a dictionary object.

    Args:
        lines (list of strings): a list of facts in the form of  A(x,y,z)@[1,2] or A@[1,2)

    Returns:
        A defaultdict object, in which the key is the predicate and the value is a dictionary (key is
        the entity and the value is a list of Interval instances) or a list of Interval instance when
        there is no entity.

    """
    lines = []
    D = defaultdict(lambda: defaultdict(list))
    if isinstance(file_or_path, list):
        lines = file_or_path[:]

    elif os.path.isdir(file_or_path):
        # It is a folder contains a set of {predicate}.csv files
        base_path = file_or_path
        for filepath in os.listdir(base_path):
            if len(filepath.split("_")) == 3:
                predicate = filepath.split("_")[1]
            else:
                predicate = filepath.split(".")[0]
            with open(os.path.join(base_path, filepath)) as file:
                reader = csv.reader(file)
                next(reader, None)
                for item in reader:
                    left_timepoint, right_timepoint = item[-2], item[-1]
                    interval = Interval(decimal.Decimal(left_timepoint), decimal.Decimal(right_timepoint), False, False)
                    # for the timestamp
                    # try:
                    #     left_timepoint = time.mktime(
                    #         datetime.datetime.strptime(left_timepoint, "%Y-%m-%d %H:%M:%S").timetuple())
                    #     right_timepoint = time.mktime(
                    #         datetime.datetime.strptime(right_timepoint, "%Y-%m-%d %H:%M:%S").timetuple())
                    #     interval = Interval(int(left_timepoint), int(right_timepoint), False, False)
                    # except:
                    #     interval = Interval(int(left_timepoint), int(right_timepoint), False, False)

                    if len(item) == 2:  # propositional
                        entity = tuple([Term("nan")])
                    else:
                        entity = tuple([Term(const) for const in item[:-2]])
                    D[predicate][entity].append(interval)
        return D

    elif isinstance(file_or_path, str) and file_or_path.endswith("csv"):
        # it is a csv file with the first column representing the predicate and the last two columns
        # representing the interval and the middle columns denotes the constants
        with open(os.path.join(file_or_path)) as file:
            reader = csv.reader(file)
            next(reader, None)
            for item in reader:
                predicate = item[0]
                left_timepoint, right_timepoint = item[-2], item[-1]
                left_timepoint = time.mktime(
                    datetime.datetime.strptime(left_timepoint, "%Y-%m-%d %H:%M:%S").timetuple())
                right_timepoint = time.mktime(
                    datetime.datetime.strptime(right_timepoint, "%Y-%m-%d %H:%M:%S").timetuple())
                interval = Interval(int(left_timepoint), int(right_timepoint), False, False)
                if len(item) == 2:  # propositional
                    entity = tuple([Term("nan")])
                else:
                    entity = tuple([Term(const) for const in item[1:-2]])
                D[predicate][entity].append(interval)
        return D

    elif isinstance(file_or_path, str) and not file_or_path.endswith("csv"):
        with open(os.path.join(file_or_path)) as file:
            for line in file:
                lines.append(line)

    else:
        raise ValueError('The input should be a file path or a list of rule string')

    for line in lines:
        line = line.strip().replace(" ","")
        if line == "":
            continue
        try:
          predicate, entity, interval = parse_str_fact(line)

        except:
            continue
        if predicate not in D:
            if entity:
               D[predicate][entity] = [interval]
            else:
               D[predicate] = [interval]
        else:
            if isinstance(D[predicate], list) and entity is not None:
                raise ValueError("One predicate can not have both entity and Null cases!")

            if not isinstance(D[predicate], list) and entity is None:
                raise ValueError("One predicate can not have both entity and Null cases!")

            if entity:
                if entity in D[predicate]:
                    D[predicate][entity].append(interval)
                else:
                    D[predicate][entity] = [interval]
            else:
                D[predicate].append(interval)


    return D


def load_program(file_or_path):
    """
    Format each string-like rule into a rule instance.

    Args:
        rules (list of strings): each string represents a rule, e.g. A(X):- Boxminus[1,2]B(X)

    Returns:
        list of rule instances
    """
    rules = []
    if isinstance(file_or_path, str):
        # it is a file path
        with open(file_or_path) as file:
            for line in file:
                line = line.replace("[-]", "Boxminus")
                line = line.replace("<+>", "Diamondplus")
                line = line.replace("<->", "Diamondminus")
                line = line.replace("[+]", "Boxplus")
                line = line.replace(" ", "")
                line = line.replace(".", "")
                rules.append(line)
    elif isinstance(file_or_path, list):
        rules = file_or_path[:]

    else:
        raise ValueError('The input should be a file path or a list of rule string')

    program = []
    for line in rules:
        rule = parse_rule(line)
        program.append(rule)

    return program
